import curses
from typing import List

def capture_tab_and_process_input(prompt: str, options: List[str]) -> str:
    stdscr = curses.initscr()
    curses.noecho()
    stdscr.keypad(True)

    user_input = ''
    cursor_position = 0

    while True:
        # stdscr.clear()
        stdscr.addstr(0, 0, prompt + user_input)
        stdscr.move(0, len(prompt) + cursor_position)
        stdscr.refresh()

        event = stdscr.getch()
        if event == ord('\t'):
            # If Tab is pressed without typing anything, show all options
            if not user_input:
                filtered_options = options
            else:
                # Process the input string so far to generate the list of options
                filtered_options = [option for option in options if option.startswith(user_input)]
            if filtered_options:
                selected_option = select_option(filtered_options, stdscr)
                if selected_option is not None:
                    user_input = selected_option
                    cursor_position = len(user_input)
        elif event == ord('\x1b'):  # Escape key
            user_input = ''
            cursor_position = 0
        elif event == curses.KEY_ENTER or event == 10:
            break  # Exit the loop and return user input
        elif event == curses.KEY_BACKSPACE or event == 127:
            if cursor_position > 0:
                user_input = user_input[:cursor_position - 1] + user_input[cursor_position:]
                cursor_position -= 1
        elif event == curses.KEY_LEFT:
            cursor_position = max(0, cursor_position - 1)
        elif event == curses.KEY_RIGHT:
            cursor_position = min(len(user_input), cursor_position + 1)
        elif event >= 32 and event <= 126:  # printable characters
            user_input = user_input[:cursor_position] + chr(event) + user_input[cursor_position:]
            cursor_position += 1

        # Show possible responses only when Tab key is pressed
        if event == ord('\t') and filtered_options:
            stdscr.addstr(1, 0, 'Possible responses: ')
            for i, option in enumerate(filtered_options):
                stdscr.addstr(option + ' ' if i < len(filtered_options) - 1 else option)
            stdscr.refresh()

    curses.endwin()  # Reset the console
    return user_input

def select_option(options: List[str], stdscr) -> str:
    selected_index = 0
    while True:
        stdscr.move(1, 0)
        stdscr.clrtoeol()
        stdscr.addstr('Possible responses: ')
        for i, option in enumerate(options):
            stdscr.addstr(option + ' ', curses.A_REVERSE if i == selected_index else curses.A_NORMAL)
        stdscr.refresh()

        event = stdscr.getch()
        if event == curses.KEY_RIGHT:
            selected_index = (selected_index + 1) % len(options)
        elif event == curses.KEY_LEFT:
            selected_index = (selected_index - 1) % len(options)
        elif event == curses.KEY_ENTER or event == 10:
            return options[selected_index]

# Example usage:
prompt = "Enter your input: "
options = ["apple", "banana", "ball", "cherry", "church", "grape", "lemon", "aphid"]
result = capture_tab_and_process_input(prompt, options)
print("You selected:", result)




str_description = """

    Some data models/schemas.

"""

from    pydantic            import BaseModel, Field
from    typing              import Optional, List, Dict
from    datetime            import datetime
from    enum                import Enum
from    pathlib             import Path

class messageType(Enum):
    """ message type/level """
    INFO    = 1
    ERROR   = 2

class loggingType(Enum):
    """ logging type """
    CONSOLE = 1
    NDJSON  = 2

class time(BaseModel):
    """A simple model that has a time string field"""
    time                                : str




import click
import click
import pudb
from pfmongo import driver, env
from argparse import Namespace
from pfmongo.models import responseModel
from pfmisc import Colors as C
import copy
from pathlib import Path
from pfmongo.models import fsModel

import pfmongo.commands.smash as smash
import pfmongo.commands.fop.cd as cd
import pfmongo.commands.docop.get as get
from pfmongo.commands.fop.pwd import dir_level
from pfmongo.commands.document import delete as doc
from pfmongo.commands.database import deleteDB as db
from pfmongo.commands.collection import deleteCol as col
import asyncio

NC = C.NO_COLOUR
GR = C.GREEN
CY = C.CYAN
PL = C.PURPLE
YL = C.YELLOW


def options_add(file: str, options: Namespace) -> Namespace:
    localoptions: Namespace = copy.deepcopy(options)
    localoptions.do = "rm"
    localoptions.file = Path(file)
    localoptions.beQuiet = True
    return localoptions


async def rm_db(options: Namespace) -> responseModel.mongodbResponse:
    resp = await db.DBdel_asModel(
        driver.settmp(
            db.options_add(str(options.file), options),
            [
                {"beQuiet": True},
                {"DBname": str(options.file)},
                {"collectionName": "void"},
            ],
        )
    )
    return resp


async def rm_collection(options: Namespace) -> responseModel.mongodbResponse:
    resp = await col.collectiondel_asModel(
        driver.settmp(
            col.options_add(str(options.file), options),
            [{"beQuiet": True}, {"collectionName": str(options.file)}],
        )
    )
    return resp


async def rm_doc(options: Namespace) -> responseModel.mongodbResponse:
    resp = await doc.deleteDo_asModel(
        driver.settmp(
            doc.options_add(str(options.file.name), options), [{"beQuiet": True}]
        )
    )
    return resp


async def rm_setName(options: Namespace) -> Namespace:
    cdResp: fsModel.cdResponse = fsModel.cdResponse()
    cdResp = await cd.toParent(options)
    if cdResp.status:
        options.file = options.file.name
    return options


async def rm_do(options: Namespace) -> responseModel.mongodbResponse:
    cwd: Path = await smash.cwd(options)
    resp: responseModel.mongodbResponse = responseModel.mongodbResponse()
    cdResp: fsModel.cdResponse = fsModel.cdResponse()
    if not (
        cdResp := await cd.toParent(
            await cd.fullPath_resolve(cd.options_add(options.file, options))
        )
    ).status:
        resp.message = cdResp.message
        return resp
    options.file = Path(options.file.name)
    match dir_level(cdResp):
        case "root":
            resp = await rm_db(options)
        case "database":
            resp = await rm_collection(options)
        case "collection":
            resp = await rm_doc(options)
        case "_":
            resp.message = "invalid directory level"
    print(resp.message)
    await cd.changeDirectory(cd.options_add(str(cwd), options))
    return resp


async def rm_asInt(options: Namespace) -> int:
    resp: responseModel.mongodbResponse = await rm_do(options)
    docDelUsage: responseModel.DocumentDeleteUsage = responseModel.DocumentDeleteUsage()
    docDelUsage.status = resp.status
    return env.response_exitCode(docDelUsage)


async def rm_asModel(options: Namespace) -> responseModel.mongodbResponse:
    return await rm_do(options)


def sync_rm_asInt(options: Namespace) -> int:
    return asyncio.run(rm_asInt(options))


def sync_rm_asModel(options: Namespace) -> responseModel.mongodbResponse:
    return asyncio.run(rm_asModel(options))


@click.command(
    cls=env.CustomCommand,
    help=f"""
delete {YL}path{NC}

SYNOPSIS
{CY}rm {YL}<path>{NC}

DESC
Delete a {YL}path{NC}. Note that the {YL}<path>{NC} can consist of a
{YL}<path>{NC} prefix specifier denoting the {YL}database{NC} and {YL}collection{NC}
to delete.

""",
)
@click.pass_context
@click.argument("path", required=False)
def rm(ctx: click.Context, path: str) -> int:
    # pudb.set_trace()
    return sync_rm_asInt(options_add(path, ctx.obj["options"]))



from pfmongo.commands import smash
from pfmongo.models import responseModel
import pfmongo.commands.fop.ls as ls
from argparse import Namespace
import pudb
import ast

from prompt_toolkit import prompt, PromptSession
from prompt_toolkit.completion import NestedCompleter
from prompt_toolkit.formatted_text import ANSI

# from prompt_toolkit.layout import Layout, HSplit, VSplit, FloatContainer, Float
from prompt_toolkit.history import InMemoryHistory
import re

history: InMemoryHistory = InMemoryHistory()


async def fileList_get(options: Namespace) -> list[str]:
    lsFiles: list[str] = []
    lsRet: int = 0
    lsResp: responseModel.mongodbResponse
    lsRet, lsResp = await ls.ls_do(ls.options_add("", "", False, options))

    if lsRet:
        return lsFiles
    try:
        lsFiles = ast.literal_eval(lsResp.message)
    except Exception as e:
        print("error: parsing this 'directory' resulted in a literal_eval exception.")
        print("There might be improperly stored objects.")
    return lsFiles


async def userInput_get(options: Namespace, **kwargs) -> str:
    noninteractive: str = ""
    files: list[str] = []
    # pudb.set_trace()
    for k, v in kwargs.items():
        if k == "noninteractive":
            noninteractive = v
        if k == "files":
            files = ast.literal_eval(v)
    if noninteractive:
        return noninteractive
    userInput: str = ""
    if not len(files):
        files = await fileList_get(options)
    sallcmds: set[str] = set(smash.fscommand)
    snofcmds: set[str] = set(smash.fscommand_noArgs)
    fcmds: list[str] = list(sallcmds.symmetric_difference(snofcmds))
    d_files: dict = {i: None for i in files}
    d_choices: dict = {i: d_files for i in fcmds}
    d_noargs: dict = {i: None for i in snofcmds}
    d_choices = {**d_choices, **d_noargs}
    completer = NestedCompleter.from_nested_dict(d_choices)
    session: PromptSession = PromptSession(history=history)
    smashprompt: str = await smash.prompt_get(options)
    userInput = await session.prompt_async(ANSI(smashprompt), completer=completer)
    return userInput



from typing import Optional, List
import readline
import rlcompleter
import pudb

class SmashCompleter(rlcompleter.Completer):
    def get_completions(self, text: str, state: int) -> Optional[str]:
        """
        Gets completion options based on the current input.

        Args:
            text: The current input text (str).
            state: The completion state (int, 0 for first completion).

        Returns:
            A completion option (str) or None if no options available.
        """

        pudb.set_trace()
        current_word = text.split()[-1]
        if not current_word:
            return None

        completions = generate_completions(current_word)
        completions = [c for c in completions if c.startswith(current_word)]

        if state < len(completions):
            return completions[state]
        else:
            return None

def initialize_readline() -> None:
    """
    Initializes the readline module for tab completion.

    Raises:
        RuntimeError: If readline is not available.
    """

    if not hasattr(readline, '__doc__'):
        raise RuntimeError("Readline module not available.")

    pudb.set_trace()
    # Disable default completion (if present)
    readline.parse_and_bind("tab: ")

    # Initialize custom completer
    completer = SmashCompleter()
    readline.set_completer(completer.get_completions)

# Replace `generate_completions` with your actual function
def generate_completions(current_word: str) -> List[str]:
    """
    Generates completion options based on the current word.

    Args:
        current_word: The current word the user is typing (str).

    Returns:
        A list of possible completions (List[str]).
    """

    # Implement your logic to generate completion options based on current_word
    # (e.g., access a database, search files, or use a dictionary)
    choices = ["apple", "aphid", "banana", "ball", "cherry", "duck"]
    return [option for option in choices if option.startswith(current_word)]

def my_repl():
    initialize_readline()
    while True:
        user_input = input("smash>>> ")
        print(user_input)
        # Process user_input here (e.g., call your command handler)

my_repl()




#!/usr/bin/env python3
#
# (c) 2023+ Fetal-Neonatal Neuroimaging & Developmental Science Center
#                   Boston Children's Hospital
#
#              http://childrenshospital.org/FNNDSC/
#                        dev@babyMRI.org
#

__version__ = "0.9.150"

from os.path import normpath
import sys

from pfmongo import pfmongo
from pfmongo import env
from pfmongo.commands import fs
from pfmongo.pfmongo import (
    options_initialize,
    parser_setup,
    parser_interpret,
    parser_JSONinterpret,
)

try:
    from . import __pkg, __version__
except:
    from pfmongo import __pkg, __version__

from argparse import ArgumentParser, Namespace
import pudb
from pfmisc import Colors as C
from typing import Any, Literal, Sequence
from pfmongo.config import settings
from pfmongo.models import dataModel
from pathlib import Path
import appdirs
import click
from click.formatting import wrap_text
import re
from pfmongo.commands import database, collection, fs, man, state, document, smash

NC = C.NO_COLOUR
GR = C.GREEN
CY = C.CYAN

options: Namespace = Namespace()


def namespace_pubattribs(space: Namespace) -> list[str]:
    attributes: list[str] = re.findall(r"\b([A-Za-z][A-Za-z_0-9]*)=", str(space))
    return attributes


def sysargv_revamp(newarg: list[str]) -> list[str]:
    """
    Small but crucial method for "revamping" the sys.argv list that
    is passed to the main click function.

    The new list is returned in addition to being set in the sys
    module.

    :param newarg: a string list of "new" arguments
    """
    executable: str = sys.argv[0]
    sys.argv = []
    sys.argv.append(executable)
    sys.argv.extend(newarg)
    return sys.argv


def main(argv: Sequence[str] = []) -> int:
    """
    The main entry point to the program. Can also be called from Python by
    passing along a "pseudo" argv list of strings (note the first element of
    the list MUST be the executable name).

    The "core" CLI args are processed here as well as some environmental checks
    are performed. If necessary, the application will "terminate" from this
    method before routing to the subcommands/groups.

    This function demonstrates a design pattern that mixes argparse and click
    parsing, hopefully leveraging something of the best of both worlds.

    :param argv: a CLI argv list of strings.
    """
    global options
    # pudb.set_trace()
    add_help: bool = False
    parser: ArgumentParser = parser_setup(
        "A client for interacting with mongo DBs", add_help
    )

    # Should we show (and exit) some man page help?
    options, extra = parser_interpret(parser, argv)

    if not env.env_statePathSet(options):
        return 10

    # This is the result of argparse
    if man.coreOptions_show(options):
        return man.man(options)

    # the click "app" interprets sys.argv, hence the revamping here
    newargv: list[str] = sysargv_revamp(extra)
    return app()


@click.group(cls=env.CustomGroup, help=pfmongo.package_description)
@click.option("--man", is_flag=True, help="show more detail about core OPTIONS")
@click.option("--version", is_flag=True, help="show program version")
@click.pass_context
def app(ctx: click.Context, man: bool, version: bool) -> int:
    """
    The main "app" -- it mostly serves as a point to route to the
    various subcommands and provide some in-line help.

    Note the --man and --version are added as options here only to
    be picked up by the "--help", but they are in fact handled by
    the parent main() method.

    :param ctx: a click "context" -- it is expanded here to transmit additional
                information to subcommands.
    :param man: a CLI --man parameter used to dispatch to the argparse handler
    :return: simply returns a 0:success or <N>:failure
    """

    global options
    if not namespace_pubattribs(options):
        options = options_initialize()
    ctx.obj = {}
    ctx.obj["options"] = options

    subcommand: str | None = click.get_current_context().invoked_subcommand

    return 0


app.add_command(database.database)
app.add_command(collection.collection)
app.add_command(fs.fs)
app.add_command(state.state)
app.add_command(document.document)
app.add_command(smash.smash)
# app.add_command(man.man)



from    pfmongo     import  env
from    pfmisc      import  Colors  as C
from    pfmongo.commands.docop import add, delete, search, showAll, get
import  click
import  pudb

NC  = C.NO_COLOUR
GR  = C.GREEN
CY  = C.CYAN

@click.group(cls = env.CustomGroup, help=f"""
commands suitable for documents: {GR}add, delete, search, showall, get{NC}

This command group provides mongo "document" level commands.

""")
def document():
    pass

document.add_command(add.add)
document.add_command(delete.delete)
document.add_command(search.search)
document.add_command(showAll.showAll)
document.add_command(get.get)



import click
from pfmongo import driver, env
from argparse import Namespace
from pfmisc import Colors as C
from pfmongo.models import responseModel
import copy
import pudb
import asyncio

NC = C.NO_COLOUR
GR = C.GREEN
CY = C.CYAN
PL = C.PURPLE
YL = C.YELLOW


def options_add(field: str, options: Namespace) -> Namespace:
    localoptions: Namespace = copy.deepcopy(options)
    localoptions.do = "listDocument"
    localoptions.argument = field
    return localoptions


async def showAll_asInt(options: Namespace) -> int:
    return await driver.run_intReturn(options)


async def showAll_asModel(options: Namespace) -> responseModel.mongodbResponse:
    return await driver.run_modelReturn(options)


def sync_showAll_asInt(options: Namespace) -> int:
    return asyncio.run(showAll_asInt(options))


def sync_showAll_asModel(options: Namespace) -> responseModel.mongodbResponse:
    return asyncio.run(showAll_asModel(options))


@click.command(
    cls=env.CustomCommand,
    help=f"""
list {PL}documents{NC} in a collection.

SYNPOSIS
{CY}showall {YL}[--field <field>]{NC}

DESC
List all the documents in a collection, showing the contents of the
document's {YL}<field>{NC} key (by default this is '{YL}_id{NC}').

""",
)
@click.option(
    "--field", type=str, help="List the value of the named field", default="_id"
)
@click.pass_context
def showAll(ctx: click.Context, field: str) -> int:
    # pudb.set_trace()
    return sync_showAll_asInt(options_add(field, ctx.obj["options"]))



from argparse import Namespace
import click
from pfmisc import Colors as C
from pfmongo import driver, env
from pfmongo.models import responseModel
from copy import deepcopy
import pudb
import asyncio

NC = C.NO_COLOUR
GR = C.LIGHT_GREEN
CY = C.CYAN
YL = C.YELLOW
PL = C.PURPLE


def options_add(database: str, options: Namespace) -> Namespace:
    localoptions: Namespace = deepcopy(options)
    localoptions.do = "connectDB"
    localoptions.argument = database
    return localoptions


async def connectTo_asInt(options: Namespace) -> int:
    return await driver.run_intReturn(options)


async def connectTo_asModel(options: Namespace) -> responseModel.mongodbResponse:
    return await driver.run_modelReturn(options)


def sync_connectTo_asInt(options: Namespace) -> int:
    return asyncio.run(connectTo_asInt(options))


def sync_connectTo_asModel(options: Namespace) -> responseModel.mongodbResponse:
    return asyncio.run(connectTo_asModel(options))


@click.command(
    cls=env.CustomCommand,
    help=f"""
associate a context with {PL}DATABASE{NC}

SYNOPSIS
{CY}connect {YL}<DATABASE>{NC}

DESC
This command connects to a mongo database called {YL}DATABASE{NC}.
A mongodb "server" can contain several "databases". A {YL}DATABASE{NC}
is the lowest (or first) level of organization in monogodb.

In order to do any operations on data, you first MUST connect to
a {PL}DATABASE{NC}.
""",
)
@click.argument("database", required=True)
@click.pass_context
def connect(ctx: click.Context, database: str) -> int:
    return sync_connectTo_asInt(options_add(database, ctx.obj["options"]))



from pfmongo import pfmongo
from pfmongo.__main__ import main

import os

from pfmongo.commands.dbop import connect, showAll, deleteDB
from pfmongo.models import responseModel

os.environ["XDG_CONFIG_HOME"] = "/tmp"


def test_database_connect_main(capsys) -> None:
    """connect to a database via click"""
    ret: int = 0
    try:
        ret = main("database connect testDB".split())
    except SystemExit:
        pass
    captured = capsys.readouterr()
    assert ret == 0
    assert 'Successfully connected database to "testDB"' in captured.out


def test_database_connect_moduleAsInt() -> None:
    """connect to a database "testDB" using a module call with an int return"""

    ret: int = connect.sync_connectTo_asInt(
        connect.options_add("testDB", pfmongo.options_initialize())
    )
    assert ret == 0


def test_database_connect_moduleAsModel() -> None:
    """connect to a database "testDB" using a module call with model return"""

    ret: responseModel.mongodbResponse = connect.sync_connectTo_asModel(
        connect.options_add("testDB", pfmongo.options_initialize())
    )
    assert "Successfully" in ret.message


def test_database_showall_main(capsys) -> None:
    """show all databases via click"""
    ret: int = 0
    try:
        ret = main("database showall".split())
    except SystemExit:
        pass
    captured = capsys.readouterr()
    assert ret == 0
    assert "admin" in captured.out


def test_database_showall_moduleAsInt() -> None:
    """show all databases using a module call with an int return"""

    ret: int = showAll.sync_showAll_asInt(
        showAll.options_add(pfmongo.options_initialize())
    )
    assert ret == 0


def test_database_showAll_moduleAsModel() -> None:
    """show all databases using a module call with model return"""

    ret: responseModel.mongodbResponse = showAll.sync_showAll_asModel(
        showAll.options_add(pfmongo.options_initialize())
    )
    assert "admin" in ret.message


def test_database_delete_main(capsys) -> None:
    """delete database via click"""
    ret: int = 0
    try:
        ret = main("database deletedb testDB".split())
    except SystemExit:
        pass
    captured = capsys.readouterr()
    assert ret == 0
    assert "Successfully" in captured.out


def test_database_delete_moduleAsInt() -> None:
    """delete databases using a module call with an int return"""

    ret: int = deleteDB.sync_DBdel_asInt(
        deleteDB.options_add("testDB", pfmongo.options_initialize())
    )
    assert ret == 0


def test_database_delete_moduleAsModel() -> None:
    """delete database using a module call with model return"""

    ret: responseModel.mongodbResponse = deleteDB.sync_DBdel_asModel(
        deleteDB.options_add("testDB", pfmongo.options_initialize())
    )

    assert "Successfully" in ret.message



from argparse import Namespace
import click
from pfmongo import driver, env
from pfmisc import Colors as C
from pfmongo.models import responseModel, fsModel
from pathlib import Path
import pudb

from pfmongo.commands.document import showAll as doc
from pfmongo.commands.state import showAll as state
from pfmongo.commands.database import showAll as mdb
from pfmongo.commands.database import connect as dbconnect
from pfmongo.commands.collection import showAll as col
from pfmongo.commands.collection import connect as colconnect
import pfmongo.commands.smash as smash
import copy
from types import SimpleNamespace
import asyncio

NC = C.NO_COLOUR
GR = C.GREEN
CY = C.CYAN
YL = C.YELLOW


def options_add(path: str, options: Namespace, create: bool = False) -> Namespace:
    localoptions: Namespace = copy.deepcopy(options)
    localoptions.beQuiet = True
    localoptions.do = "cd"
    localoptions.cd = SimpleNamespace(**{"path": Path(path), "create": create})
    return localoptions


def path_to_dbCol(options: Namespace) -> tuple:
    path: Path = options.cd.path
    db: str = ""
    collection: str = ""
    match len(path.parts):
        case _ if len(path.parts) == 2:
            (root, db) = path.parts
        case _ if len(path.parts) > 2:
            (root, db, collection) = path.parts[:3]
    return (db, collection)


async def toParent(options: Namespace) -> fsModel.cdResponse:
    return await changeDirectory(options_add(options.cd.path.parent, options))


async def toDir(options: Namespace) -> fsModel.cdResponse:
    (db, collection) = path_to_dbCol(options)
    return await changeDirectory(options_add(f"/{db}/{collection}", options))


async def fullPath_resolve(options: Namespace) -> Namespace:
    async def navigate_cd() -> None:
        ups: str = "../"
        path: Path = await smash.cwd(options)
        cd = str(options.cd.path)
        if cd.startswith(ups):
            if not cd.endswith("/"):
                cd += "/"
            levels_up: int = cd.count(ups)
            try:
                path = path.parents[levels_up - 1] if levels_up > 1 else path.parent
            except Exception:
                path = Path("/")
            cd = cd[len(ups) * levels_up :]
        options.cd.path = path / cd

    def filter_toFirstStarInPath(path: Path) -> Path:
        index = next(
            (i for i, part in enumerate(path.parts) if "*" in part), len(path.parts)
        )
        return Path(*path.parts[:index])

    options.cd.path = filter_toFirstStarInPath(options.cd.path)
    if options.cd.path.is_absolute():
        return options
    if options.cd.path == Path("."):
        options.cd.path = await smash.cwd(options)
    elif options.cd.path == Path(".."):
        path: Path = await smash.cwd(options)
        options.cd.path = path.parent
    elif await smash.cwd(options) == Path("/") and not str(options.cd.path).startswith(
        ".."
    ):
        options.cd.path = Path("/") / options.cd.path
    else:
        await navigate_cd()
    return options


async def db_isListed(db: str, options: Namespace) -> fsModel.cdResponse:
    fsPath: fsModel.cdResponse = fsModel.cdResponse()
    allDB: responseModel.mongodbResponse = await mdb.showAll_asModel(
        driver.settmp(options, [{"do": "showAllDB"}])
    )
    if db not in allDB.message:
        fsPath.error = f"database '{db}' not in mongo"
        fsPath.status = False
        fsPath.code = 2
        fsPath.state["database"] = "void"
    else:
        fsPath.state["database"] = "exists"
    return fsPath


async def db_connect(db: str, options: Namespace) -> fsModel.cdResponse:
    fsPath: fsModel.cdResponse = fsModel.cdResponse()
    fsPath.path = options.cd.path
    if dbConnectFail := await dbconnect.connectTo_asInt(
        dbconnect.options_add(db, options)
    ):
        fsPath.error = f"could not connect to database '{db}'"
        fsPath.status = False
        fsPath.code = 3
        return fsPath
    return fsPath


async def col_isListed(
    db: str, collection: str, options: Namespace
) -> fsModel.cdResponse:
    fsPath: fsModel.cdResponse = fsModel.cdResponse()
    allCollections: responseModel.mongodbResponse = await col.showAll_asModel(
        driver.settmp(options, [{"do": "showAllCollections"}])
    )
    if collection not in allCollections.message:
        fsPath.error = f"collection '{collection}' not in database '{db}'"
        fsPath.status = False
        fsPath.code = 4
        fsPath.state["collection"] = "void"
    else:
        fsPath.state["collection"] = "exists"
    return fsPath


async def col_connect(
    db: str, collection: str, options: Namespace
) -> fsModel.cdResponse:
    fsPath: fsModel.cdResponse = fsModel.cdResponse()
    fsPath.path = options.cd.path
    if colCollectFail := await colconnect.connectTo_asInt(
        colconnect.options_add(collection, options)
    ):
        fsPath.error = (
            f"could not connect to collection '{collection}' in databse '{db}'"
        )
        fsPath.status = False
        fsPath.code = 5
    return fsPath


def is_root(options: Namespace) -> bool:
    b_ret: bool = False
    if not len(options.cd.path.parents):
        env.DBname_stateFileSave(options, "")
        env.collectionName_stateFileSave(options, "")
        b_ret = True
    return b_ret


def is_path_too_long(options) -> fsModel.cdResponse:
    fsPath: fsModel.cdResponse = fsModel.cdResponse()
    if len(options.cd.path.parents) > 2:
        fsPath.error = f"{options.cd.path}: path too long"
        fsPath.status = False
        fsPath.code = 1
    return fsPath


async def path_connect(options: Namespace) -> fsModel.cdResponse:
    fsPath: fsModel.cdResponse = fsModel.cdResponse()
    connect: fsModel.cdResponse = fsModel.cdResponse()
    connect.path = options.cd.path
    if (fsPath := is_path_too_long(options)).code:
        return fsPath
    if is_root(options):
        return connect
    (db, collection) = path_to_dbCol(options)
    if db and not (fsPath := await db_isListed(db, options)).status:
        if not options.cd.create:
            return fsPath
    connect.state["database"] = fsPath.state["database"]
    if db and not (fsPath := await db_connect(db, options)).status:
        return fsPath
    if (
        collection
        and not (fsPath := await col_isListed(db, collection, options)).status
    ):
        if not options.cd.create:
            return fsPath
    connect.state["collection"] = fsPath.state["collection"]
    if collection and not (fsPath := await col_connect(db, collection, options)).status:
        return fsPath
    connect.status = True
    connect.message = f"successfully connected path to {connect.path}"
    connect.code = 0
    env.DBname_stateFileSave(options, db)
    env.collectionName_stateFileSave(options, collection)
    return connect


def path_lengthOK(options: Namespace) -> bool:
    return True if len(options.cd.path.parents.parts == 2) else False


async def changeDirectory(options: Namespace) -> fsModel.cdResponse:
    options = await fullPath_resolve(options)
    cdResponse: fsModel.cdResponse = fsModel.cdResponse()
    if (cdResponse := await path_connect(options)).code:
        print(cdResponse.error)
    return cdResponse


def sync_changeDirectory(options: Namespace) -> fsModel.cdResponse:
    return asyncio.run(changeDirectory(options))


@click.command(
    cls=env.CustomCommand,
    help=f"""
change {YL}directory{NC}

SYNOPSIS
{CY}cd {YL}[--create] <path>{NC}

DESC
The {YL}cd{NC} command "changes directory" within a mongodb between a
{CY}<database>{NC} level at the root `/` and a {CY}<collection>{NC} within that
{CY}<database>{NC}. Since there are only ever two "directories" in a
mongodb, trees are not especially nested.

Note that the underlying behaviour is subtly different to a standard
filesystem, and this 'cd' can also "create" a path/directory with if called
with {YL}--create{NC}.

""",
)
@click.argument("path", required=True)
@click.option(
    "--create", is_flag=True, help="If set, also create this path if it does not exist."
)
@click.pass_context
def cd(ctx: click.Context, path: str, create: bool) -> int:
    # pudb.set_trace()
    return sync_changeDirectory(options_add(path, ctx.obj["options"], create)).code



import click
import click
import pudb
from typing import Any
import readline
import sys
from tabulate import tabulate
from pfmongo import pfmongo, __main__
from pfmongo import driver, env
from argparse import Namespace
from pfmisc import Colors as C
from pfmongo.models import responseModel, fsModel
from typing import cast
from argparse import Namespace
from pfmongo.commands.state import showAll as state
from click.testing import CliRunner
from pathlib import Path
import ast
from fortune import fortune
import copy
import pfmongo.commands.fop.prompt as prompt
from pfmongo.config import settings
import subprocess
from typing import Optional, Callable, Union
import asyncio
from asyncio import get_event_loop
from pydantic import BaseModel

# from    ansi2html               import Ansi2HTMLConverter
from pfmongo.commands.slib import tabc

NC = C.NO_COLOUR
GR = C.GREEN
CY = C.CYAN
YL = C.YELLOW

fscommand: list = [
    "sg",
    "ls",
    "cat",
    "rm",
    "cd",
    "mkcd",
    "imp",
    "exp",
    "prompt",
    "pwd",
    "quit",
    "exit",
    "fortune",
    "banner",
]

fscommand_noArgs: list = ["prompt", "pwd", "quit", "exit"]


def pipe_split(command: str) -> list:
    parts: list[str] = command.split("|", 1)
    return parts


def smash_output(command: str) -> str:
    output: str = meta_parse(command)
    if not output:
        ret = CliRunner().invoke(__main__.app, command.split(), color=True)
        output = ret.output
    return output


class ClickResult(BaseModel):
    exit_code: int
    output: str
    # Add other fields if needed


async def invoke_click_async(app, args: list[str], **kwargs) -> ClickResult:
    """
    Asynchronously invoke a Click command.

    :param app: The Click application.
    :param args: List of command arguments.
    :param kwargs: Additional keyword arguments for CliRunner.invoke.
    :return: A ClickResult instance with the command's result.
    """
    loop = asyncio.get_running_loop()
    runner = CliRunner()
    result = await loop.run_in_executor(
        None, lambda: runner.invoke(app, args, **kwargs)
    )
    return ClickResult(exit_code=result.exit_code, output=result.output)


async def smash_output_async(command: str) -> str:
    """
    Asynchronous version of smash_output.

    :param command: The command string to process.
    :return: The output of the command.
    """
    output: str = meta_parse(command)
    if not output:
        result: ClickResult = await invoke_click_async(
            __main__.app, command.split(), color=True
        )
        output = result.output
    return output


async def smash_execute_async(
    command: str,
    f: Optional[Callable[[str, list[str]], subprocess.CompletedProcess]] = None,
) -> Union[bytes, str]:
    """
    Asynchronous version of smash_execute.

    :param command: The command string to process.
    :param f: Optional callable to handle piped commands.
    :return: Result of the command execution.
    """
    cmdpart: list[str] = pipe_split(command)
    smash_ret: str = await smash_output_async(cmdpart[0])
    result: str = smash_ret

    if len(cmdpart) > 1 and f:
        # Handle potential blocking operation in a separate thread
        loop = asyncio.get_running_loop()
        process: subprocess.CompletedProcess = await loop.run_in_executor(
            None, lambda: f(smash_ret, cmdpart)
        )
        result = f"exec: '{process.args}', returncode: {process.returncode}"
        result = ""

    return result


def smash_execute(
    command: str,
    f: Optional[Callable[[str, list[str]], subprocess.CompletedProcess]] = None,
) -> Union[bytes, str]:
    cmdpart: list = pipe_split(command)
    smash_ret: str = smash_output(cmdpart[0])
    result: str = smash_ret
    if len(cmdpart) > 1 and f:
        process: subprocess.CompletedProcess = f(smash_ret, cmdpart)
        result = f"exec: '{process.args}', returncode: {process.returncode}"
        result = ""
    return result


async def smash_executeAsync(
    command: str,
    f: Optional[Callable[[str, list[str]], subprocess.CompletedProcess]] = None,
) -> Union[bytes, str]:
    cmdpart: list = pipe_split(command)
    smash_ret: str = smash_output(cmdpart[0])
    result: str = smash_ret
    if len(cmdpart) > 1 and f:
        process: subprocess.CompletedProcess = f(smash_ret, cmdpart)
        result = f"exec: '{process.args}', returncode: {process.returncode}"
        result = ""
    return result


def pipe_handler(previous_input: str, cmdpart: list) -> subprocess.CompletedProcess:
    cmds = [c.strip() for c in cmdpart]
    shell_command = "|".join(cmds[1:])
    result: subprocess.CompletedProcess = subprocess.run(
        shell_command, input=previous_input, shell=True, capture_output=False, text=True
    )
    # converter:Ansi2HTMLConverter    = Ansi2HTMLConverter()
    # output:str                      = converter.convert(result.stdout)
    return result


def command_parse(command: str) -> str:
    fscall: list = [s for s in fscommand if command.lower().startswith(s)]
    if fscall:
        command = f"fs {command}"
    if command == "help":
        command = "fs --help"
    return command


async def state_getModel(options: Namespace) -> responseModel.mongodbResponse:
    return await state.showAll_asModel(driver.settmp(options, [{"beQuiet": True}]))


async def cwd(options: Namespace) -> Path:
    model: responseModel.mongodbResponse = await state_getModel(options)
    if model.message == "/":
        return Path("/")
    else:
        return Path("/" + model.message)


async def prompt_get(options: Namespace) -> str:
    promptDo: fsModel.cdResponse = await prompt.prompt_do(
        await prompt.options_add(options)
    )
    pathColor: str = promptDo.message
    return f"{CY}({settings.mongosettings.MD_sessionUser}@smash){NC}{pathColor}$>"


async def command_get(options: Namespace, **kwargs) -> str:
    userInput: str = await tabc.userInput_get(options, **kwargs)
    fscmd: str = f"{userInput}".strip()
    # pudb.set_trace()
    return fscmd


def sync_command_get(options: Namespace, **kwargs) -> str:
    return asyncio.run(command_get(options, **kwargs))


def meta_parse(command: str) -> str:
    output: str = ""
    if "quit" in command.lower() or "exit" in command.lower():
        sys.exit(0)
    if "banner" in command.lower():
        output = introbanner_generate()
    if "fortune" in command.lower():
        output = env.tabulate_message(fortune(), f"{YL}fortune{NC}")
    return output


def introbanner_generate() -> str:
    title: str = (
        f"{CY}s{YL}imple pf{CY}m{YL}ongo{NC} {CY}a{YL}pplication {CY}sh{YL}ell{NC}"
    )
    banner: str = f"""
{CY}░░░░░░░░░░░░░░░░░░░░░░░░░░░░░░░░░░░░░░░░░░░░░░░░{NC}
{CY}░███████╗███╗░░░███╗░█████╗░███████╗██╗░░██╗██╗░{NC}
{CY}░██╔════╝████╗░████║██╔══██╗██╔════╝██║░░██║██║░{NC}
{CY}░███████╗██╔████╔██║███████║███████╗███████║██║░{NC}
{CY}░╚════██║██║╚██╔╝██║██╔══██║╚════██║██╔══██║╚═╝░{NC}
{CY}░███████║██║░╚═╝░██║██║░░██║███████║██║░░██║██╗░{NC}
{CY}░╚══════╝╚═╝░░░░░╚═╝╚═╝░░╚═╝╚══════╝╚═╝░░╚═╝╚═╝░{NC}
{CY}░░░░░░░░░░░░░░░░░░░░░░░░░░░░░░░░░░░░░░░░░░░░░░░░{NC}
"""
    intro: str = f"""
Welcome to {CY}smash{NC}, a simple shell to {YL}pfmongo{NC} that
allows you to directly call various subcommands.

If you really want to {CY}smash{NC} things up, the shell allows
you to call the {YL}fs{NC} subcommands without the {YL}fs{NC}, giving
the impression of being in a more standard un*x land.

Some useful commands:
 ▙▄ {YL}--help{NC} for a list of {YL}all{NC} commands.
 ▙▄ {YL}banner{NC} to see this amazing banner again.
 ▙▄ {YL}fortune{NC} since every shell needs this.
 ▙▄ {YL}help{NC} for a list of {YL}fs{NC} commands.
 ▙▄ {YL}quit{NC} or {YL}exit{NC} to return to the system.

Enjoy your stay and please remain on the trails!
Oh, and don't feed the wildlife. It's not good
for them.

Have a {CY}smash{NC}ing good time!
"""
    return env.tabulate_message(banner + intro, title)


@click.command(
    cls=env.CustomCommand,
    help="""
shell interface for running commands

An extremely "simple" pfmongo {YL}shell{NC}. Run commands from a shell-esque
interface that harkens back to the days of /bin/ash!

""",
)
@click.option("--prompt", is_flag=True, help="If set, print the CFS cwd as prompt")
@click.pass_context
def smash(ctx: click.Context, prompt) -> None:
    print(introbanner_generate())
    options: Namespace = ctx.obj["options"]
    loop: asyncio.AbstractEventLoop = get_event_loop()
    while True:
        command_str: str = loop.run_until_complete(command_get(options))
        print(smash_execute(command_parse(command_str), pipe_handler))



from    argparse        import Namespace
from    pfmongo         import driver, pfmongo
from    pfmongo.pfmongo import package_CLIfull, package_argsSynopsisFull, package_argSynopsisSelf
import  pudb
from    pfmisc          import  Colors as C
import  os, sys

try:
    from    .               import __pkg, __version__
except:
    from pfmongo            import __pkg, __version__

NC  = C.NO_COLOUR


str_title:str = r'''

           $$$$$$\
          $$  __$$\
 $$$$$$\  $$ /  \__|$$$$$$\$$$$\   $$$$$$\  $$$$$$$\   $$$$$$\   $$$$$$\
$$  __$$\ $$$$\     $$  _$$  _$$\ $$  __$$\ $$  __$$\ $$  __$$\ $$  __$$\
$$ /  $$ |$$  _|    $$ / $$ / $$ |$$ /  $$ |$$ |  $$ |$$ /  $$ |$$ /  $$ |
$$ |  $$ |$$ |      $$ | $$ | $$ |$$ |  $$ |$$ |  $$ |$$ |  $$ |$$ |  $$ |
$$$$$$$  |$$ |      $$ | $$ | $$ |\$$$$$$  |$$ |  $$ |\$$$$$$$ |\$$$$$$  |
$$  ____/ \__|      \__| \__| \__| \______/ \__|  \__| \____$$ | \______/
$$ |                                                  $$\   $$ |
$$ |                                                  \$$$$$$  |
\__|                                                   \______/
'''


str_heading:str = f"""
                        python (pf) monogodb client and module

"""

def synopsis_print(summary:bool = False) -> int:
    scriptName:str          = os.path.basename(sys.argv[0])
    print(C.CYAN + '''
    NAME
        ''', end = '' + NC)
    print(scriptName)
    print(C.CYAN + '''
    CORE SYNPOSIS
        ''' + NC, end = '')
    print(scriptName + pfmongo.package_CLIfull)
    print(C.CYAN + '''
    DESCRIPTION ''' + NC, end = '')
    print(pfmongo.package_coreDescription)

    if summary: return 1
    print(C.CYAN + '''
    ARGS''' + NC, end="")
    print(pfmongo.package_argsSynopsisFull)
    return 2

def coreOptions_show(options:Namespace) -> bool:
    ret:bool    = False
    if options.man or options.version:
        ret     = True
    return ret

def versionAndName_print() -> int:
    print("Name:    ", end="")
    print(C.LIGHT_GREEN + f'{__pkg.name}' + NC)
    print("Version: ", end="")
    print(C.LIGHT_CYAN + f'{__version__}\n')
    return 1

def coreManpage_print() -> int:
    print(f'''
          {C.GREEN}{str_title}
          {str_heading}{NC}
          ''')
    summary:bool    = False
    return synopsis_print(summary)

def man(options:Namespace) -> int:
    #pudb.set_trace()
    if options.version:
        return versionAndName_print()
    if options.man:
        return coreManpage_print()
    return 0



import builtins

from _pytest.capture import capsys
from pfmongo import pfmongo
from pfmongo.__main__ import main
import os

os.environ["XDG_CONFIG_HOME"] = "/tmp"
import pytest

from pfmongo.commands.dbop import connect, showAll
from pfmongo.models import responseModel


def CLIcall_parseForStringAndRet(
    capsys, cli: str, contains: str, exitCode: int
) -> None:
    print(f"Testing {cli}")

    ret: int = 0
    try:
        ret = main(cli.split())
    except SystemExit:
        raise
    ret = main(cli.split())
    captured = capsys.readouterr()
    assert ret == exitCode
    assert contains in captured.out


def test_main_manCore(capsys) -> None:
    """core man page"""
    CLIcall_parseForStringAndRet(capsys, "--man", "--useDB <DBname>", 2)


def test_main_version(capsys) -> None:
    """version CLI reporting"""
    CLIcall_parseForStringAndRet(capsys, "--version", "Name", 1)


def test_imp_help(capsys) -> None:
    with pytest.raises(builtins.SystemExit) as exit_info:
        CLIcall_parseForStringAndRet(capsys, "fs imp --help", "imports", 0)

    assert exit_info.value.args == (0,)



import readline
from typing import Optional
import pudb

fruits: list[str] = ["apple", "aphid", "berry", "balloon", "banana", "ball", "cherry", "cherub", "duck"]
colors: list[str] = ["red", "darkred", "green", "darkgreen"]

current_choices: list[str] = colors
first_word:str      = ""

def completer(text: str, state: int) -> Optional[str]:
    global current_choices, first_word
#    pudb.set_trace()
    options: list[str] = [choice for choice in current_choices if choice.startswith(text)]
    if state < len(options):
        for index, item in enumerate(options):
            if text == item:
                current_choices     = fruits
                state = index
                if not first_word:
                    first_word = text
                break  # Exit the loop after finding the first match

        return options[state]
    else:
        return None

readline.set_completer(completer)
readline.parse_and_bind("tab: complete")

def get_choice(prompt: str) -> str:
    global current_choices, first_word
    current_choices = colors
    first_word      = ""
    while True:
        user_input: str = input(prompt).strip()
        if current_choices == fruits:
            user_input = user_input.split()[1]
        if not user_input:  # If no input, show all choices
            print("Possible choices:", ", ".join(current_choices))
            continue
        matching_choices: list[str] = [choice for choice in current_choices if choice.startswith(user_input)]
        if len(matching_choices) == 1:  # If only one choice left, return it
            return matching_choices[0]
        elif len(matching_choices) > 1:  # If multiple choices, prompt again with completion
            completion: Optional[str] = completer(user_input, 0)
            if completion is not None:
                user_input += completion[len(user_input):]  # Update user_input with completion
                readline.redisplay()
        else:
            print("No matches found. Please try again.")
            continue

# Example REPL usage:
while True:
    prompt: str = "Enter your choice: "
    choice: str = get_choice(prompt)
    print("You selected:", first_word + ' ' + choice)





import click
import pudb
from pfmongo import driver, env
import pfmongo.pfmongo
from argparse import Namespace
import json
from pfmisc import Colors as C
from pfmongo.config import settings
from pfmongo.models import responseModel
from typing import Tuple, cast, Callable
from pfmongo.commands.docop import add
from pathlib import Path
import copy
import asyncio

NC = C.NO_COLOUR
GR = C.GREEN
CY = C.CYAN
PL = C.PURPLE
YL = C.YELLOW


def filename_get(path: str) -> str:
    filename: str = Path(path).name
    return filename


def options_add(file: str, options: Namespace) -> Namespace:
    localoptions: Namespace = copy.deepcopy(copy.deepcopy(options))
    localoptions.beQuiet = True
    localoptions.do = "import"
    localoptions.argument = {"file": file}
    return localoptions


async def imp_do(options: Namespace) -> responseModel.mongodbResponse:
    imp: responseModel.mongodbResponse = await add.documentAdd_asModel(
        add.options_add(
            options.argument["file"], filename_get(options.argument["file"]), options
        )
    )
    return imp


async def imp_asInt(options: Namespace) -> int:
    imp: responseModel.mongodbResponse = await imp_do(options)
    print(imp.message)
    return 0


def sync_imp_asInt(options: Namespace) -> int:
    return asyncio.run(imp_asInt(options))


@click.command(
    cls=env.CustomCommand,
    help=f"""
import a {CY}file{NC} as a {YL}document{NC}

SYNOPSIS
{CY}imp {YL}<file>{NC}

DESC
Import a {YL}file{NC} from the host system into the mongodb as a {YL}document{NC}.

The {YL}file{NC} can have a "path" prefix, denoting a database and collection to
use for the import.

""",
)
@click.argument("file", required=True)
@click.pass_context
def imp(ctx: click.Context, file: str) -> int:
    # pudb.set_trace()
    return sync_imp_asInt(options_add(file, ctx.obj["options"]))



from argparse import Namespace
import click
from pfmisc import Colors as C
from pfmongo import driver, env
from pfmongo.models import responseModel
from pfmongo.config import settings
from typing import Tuple
import copy
import asyncio

NC = C.NO_COLOUR
GR = C.LIGHT_GREEN
CY = C.CYAN
PL = C.PURPLE
YL = C.YELLOW


def options_add(collection: str, options: Namespace) -> Namespace:
    localoptions = copy.deepcopy(options)
    localoptions.do = "connectCollection"
    localoptions.argument = collection
    return localoptions


def is_shadowCollection(collection: str) -> Tuple[str, bool]:
    isShadow: bool = (
        True if collection.endswith(settings.mongosettings.flattenSuffix) else False
    )
    return collection, isShadow


async def baseCollection_getAndConnect(options: Namespace) -> Namespace:
    localoptions = copy.deepcopy(options)
    localoptions.collectionName = env.collectionName_get(localoptions)
    currentCol: str = env.collectionName_get(options)
    if currentCol.endswith(settings.mongosettings.flattenSuffix):
        currentCol = currentCol.rstrip(settings.mongosettings.flattenSuffix)
        localoptions = await collection_connect(currentCol, options)
    return localoptions


async def shadowCollection_getAndConnect(options: Namespace) -> Namespace:
    localoptions = copy.deepcopy(options)
    localoptions.collectionName = env.collectionName_get(localoptions)
    shadowCol: str = ""
    if not (currentCol := is_shadowCollection(env.collectionName_get(options)))[1]:
        shadowSuffix: str = settings.mongosettings.flattenSuffix
        shadowCol = currentCol[0] + shadowSuffix
        localoptions = await collection_connect(shadowCol, options)
    return localoptions


async def collection_connect(collection: str, options: Namespace) -> Namespace:
    """Returns a copy of the options with the following
    state changes:

        1. 'runRet' contains the int result of doing a connection
                    to the <collection>. This run is also performed
                    on its own options copy.
        2. 'collectionName' in the original copy is updated to the
                            <collection>.

    Note the original options passed into this method is *unchanged*!
    """
    return driver.settmp(
        options,
        [
            {
                "runRet": await driver.run_intReturn(
                    driver.settmp(
                        options, [{"do": "connectCollection"}, {"argument": collection}]
                    )
                )
            },
            {"collectionName": collection},
        ],
    )


async def connectTo_asInt(options: Namespace) -> int:
    return await driver.run_intReturn(options)


async def connectTo_asModel(options: Namespace) -> responseModel.mongodbResponse:
    return await driver.run_modelReturn(options)


def sync_connectTo_asInt(options: Namespace) -> int:
    return asyncio.run(connectTo_asInt(options))


def sync_connectTo_asModel(options: Namespace) -> responseModel.mongodbResponse:
    return asyncio.run(connectTo_asModel(options))


@click.command(
    cls=env.CustomCommand,
    help=f"""
associate context with {PL}COLLECTION{NC}

SYNOPSIS
{CY}connect {YL}<COLLECTION>{NC}

DESC
This command connects to a mongo collection called {YL}COLLECTION{NC}
within a mongo database.

A mongodb "server" can contain several "databases", each of which
contains several "collections".

A {YL}COLLECTION{NC} is the second level of organization in a monogdb.

""",
)
@click.argument("collection", required=True)
@click.pass_context
def connect(ctx: click.Context, collection: str) -> int:
    return sync_connectTo_asInt(options_add(collection, ctx.obj["options"]))



from pathlib import Path
from argparse import Namespace
from typing import Optional
from pfmongo import env
from pfmisc import Colors as C
import click
import pudb

from pfmongo.commands.fop import cd, ls, cat, mkcd, imp, exp, pwd, prompt, rm, sg

NC = C.NO_COLOUR
GR = C.GREEN
CY = C.CYAN


def root(options: Namespace) -> Path:
    root: Path = Path()
    RFS: Path = options.thisSession
    splitPoint: str = "_MONGO_"
    si: Optional[int] = next(
        (i for i, segment in enumerate(RFS.parts) if splitPoint in segment), None
    )
    if si is not None:
        root = Path(*list(RFS.parts)[: si + 1])
    return root


@click.group(
    cls=env.CustomGroup,
    help=f"""
commands suitable for "{GR}file system{NC}" abstractions

This command group uses file system (FS) "commands" in the context of a mongodb
allowing for an FS-modeled interface.

""",
)
def fs():
    pass


fs.add_command(ls.ls)
fs.add_command(cat.cat)
fs.add_command(cd.cd)
fs.add_command(mkcd.mkcd)
fs.add_command(imp.imp)
fs.add_command(exp.exp)
fs.add_command(pwd.pwd)
fs.add_command(prompt.prompt)
fs.add_command(rm.rm)
fs.add_command(sg.sg)



import click
from pfmongo import driver, env
from argparse import Namespace
from pfmisc import Colors as C
from pfmongo.models import fsModel, responseModel
from pathlib import Path
import ast
import pudb
import copy
import re
import fnmatch
import pfmongo.commands.smash as smash
import pfmongo.commands.fop.cd as cd
from pfmongo.commands.document import showAll as doc
from pfmongo.commands.database import showAll as db
from pfmongo.commands.collection import showAll as col
from pfmongo.commands.fop.pwd import pwd_level
import asyncio
from types import SimpleNamespace
import fontawesome as fa
import shutil

NC = C.NO_COLOUR
GR = C.GREEN
CY = C.CYAN
PL = C.PURPLE
YL = C.YELLOW


def options_add(path: str, attribs: str, long: bool, options: Namespace) -> Namespace:
    localoptions = copy.deepcopy(options)
    localoptions.do = "ls"
    localoptions.beQuiet = True
    if not path:
        path = "."
    localoptions.argument = SimpleNamespace(
        **{"path": Path(path), "attribs": attribs, "long": long}
    )
    return localoptions


def objectSymbol_resolve(resp: responseModel.mongodbResponse) -> str:
    symbol: str = ""
    match type(resp.response["connect"]):
        case responseModel.showAllDBusage:
            symbol = fa.icons["database"]
        case responseModel.showAllcollectionsUsage:
            symbol = fa.icons["table"]
        case responseModel.DocumentListUsage:
            symbol = fa.icons["file"]
        case _:
            symbol = fa.icons["file"]
    return symbol + " "


def resp_process(resp: responseModel.mongodbResponse) -> str:
    rstr: str = ""
    try:
        file: list = ast.literal_eval(resp.message)
        for f in file:
            rstr += objectSymbol_resolve(resp) + f + "  "
    except Exception as e:
        pass
    return rstr


async def ls_db(options: Namespace) -> responseModel.mongodbResponse:
    resp = await db.showAll_asModel(
        driver.settmp(db.options_add(options), [{"beQuiet": True}])
    )
    return resp


async def ls_collection(options: Namespace) -> responseModel.mongodbResponse:
    resp = col.showAll_asModel(
        driver.settmp(col.options_add(options), [{"beQuiet": True}])
    )
    return await resp


def file_filter(
    resp: responseModel.mongodbResponse, options: Namespace
) -> responseModel.mongodbResponse:
    if options.argument.path != Path("."):
        files: list[str] = eval(resp.message)
        filtered: list[str] = [
            e for e in files if fnmatch.fnmatch(e, options.argument.path)
        ]
        resp.message = "[" + ", ".join([f"'{e}'" for e in filtered]) + "]"
    return resp


async def ls_doc(options: Namespace) -> responseModel.mongodbResponse:
    resp = await doc.showAll_asModel(
        driver.settmp(doc.options_add("_id", options), [{"beQuiet": True}])
    )
    if resp.message:
        resp.message = ls_msgParse(resp.message)
    return resp


def ls_msgParse(lstr: str) -> str:
    return re.sub(r"ObjectId\('([\w]+)'\)", r"'\1'", lstr)


async def ls_fargsUpdate(options: Namespace) -> Namespace:
    localoptions: Namespace = copy.deepcopy(options)
    path: Path = await smash.cwd(options)
    cwdSet: set[str] = set(path.parts[-1:])
    argSet: set[str] = set(options.argument.path.parts[-1:])
    try:
        localoptions.argument.path = (argSet - cwdSet).pop()
    except KeyError:
        localoptions.argument.path = Path(".")
    return localoptions


async def ls_do(options: Namespace) -> tuple[int, responseModel.mongodbResponse]:
    cwd: Path = await smash.cwd(options)
    resp: responseModel.mongodbResponse = responseModel.mongodbResponse()
    cdResp: fsModel.cdResponse = fsModel.cdResponse()
    # pudb.set_trace()
    # cd.changeDirectory(cd.options_add(options.argument.path, options))
    if not (
        cdResp := await cd.toDir(
            await cd.fullPath_resolve(cd.options_add(options.argument.path, options))
        )
    ).status:
        resp.message = cdResp.message + " " + cdResp.error
        return 1, resp

    loptions = await ls_fargsUpdate(options)
    match await pwd_level(loptions):
        case "root":
            resp = await ls_db(loptions)
        case "database":
            resp = await ls_collection(loptions)
        case "collection":
            resp = await ls_doc(loptions)
        case "_":
            resp = await ls_db(loptions)
    file_filter(resp, loptions)
    if not options.beQuiet:
        resp_process(resp)
    ret: int = 0
    if not resp.message:
        ret = 1
    await cd.changeDirectory(cd.options_add(str(cwd), options))
    return ret, resp


async def ls_asInt(options: Namespace) -> int:
    ret, resp = await ls_do(options)
    return ret


async def ls_asModel(options: Namespace) -> responseModel.mongodbResponse:
    ret, resp = await ls_do(options)
    return resp


def sync_ls_asInt(options: Namespace) -> int:
    return asyncio.run(ls_asInt(options))


def sync_ls_asModel(options: Namespace) -> responseModel.mongodbResponse:
    return asyncio.run(ls_asModel(options))


def terminal_getWidth() -> int:
    return shutil.get_terminal_size().columns


def columns_format(items: list[str], terminal_width: int) -> list[str]:
    """
    Format the given items into columns that fit the terminal width.

    Args:
        items (list[str]): List of strings to format.
        terminal_width (int): Width of the terminal in characters.

    Returns:
        list[str]: List of formatted strings, each representing a row.
    """
    max_item_width: int = max(len(item) for item in items)
    column_width: int = max_item_width + 2  # Add 2 for spacing

    num_columns: int = max(1, terminal_width // column_width)
    num_rows: int = (len(items) + num_columns - 1) // num_columns

    rows: list[str] = []
    for row in range(num_rows):
        row_items: list[str] = items[row::num_rows]
        formatted_row: str = "".join(
            item.ljust(column_width) for item in row_items
        ).rstrip()
        rows.append(formatted_row)

    return rows


def columns_print(items: list[str]) -> None:
    """
    Print the given items in columns that fit the terminal width.

    Args:
        items (list[str]): List of strings to print.
    """
    terminal_width: int = terminal_getWidth()
    formatted_rows: list[str] = columns_format(items, terminal_width)

    for row in formatted_rows:
        print(row)


def filenames_parse(data: str) -> list[str]:
    """
    Parse the data string into a list of filenames.

    Args:
        data (str): The input data string.

    Returns:
        list[str]: List of filenames, each consisting of a type letter, a space, and a string.
    """
    lines: list[str] = data.strip().split("\n")
    filenames: list[str] = [line.strip() for line in lines if line.strip()]
    return filenames


@click.command(
    cls=env.CustomCommand,
    help=f"""
list {YL}path{NC}

SYNOPSIS
{CY}ls {YL}[--attribs {GR}<attriblist>{YL}] [--raw] [--human] [--long] [<path>]{NC}

ARGS
{YL}--long{NC}
This flag triggers a detailed listing, showing analogues to the document
or "file" {CY}owner{NC}, {CY}size{NC}, and creation {CY}date{NC}.

This assumes that the document entry has these fields encoded, which 
is only true for files uploaded using {YL}pfmongo{NC}.

{YL}--raw{NC}
If specified, show the raw response.

{YL}--human{NC}
If specified with {YL}--long{NC} show any sizes in human friendly terms.

{YL}--attribs {GR}<attriblist>{NC}
If specified with {YL}--long{NC} show the passed attributes, specified as 
a comma separated list.

DESCRIPTION
This command lists the objects (files and directories) that are at a given
path. This path can be a directory, in which case possibly multiple objects
are listed, or it can be a single file in which case information about that
single file is listed.


""",
)
@click.argument("path", required=False)
@click.option(
    "--attribs",
    required=False,
    help="A comma separated list of file attributes to return/print",
)
@click.option(
    "--raw",
    is_flag=True,
    help="print the actual returned response without cleanup",
)
@click.option(
    "--human",
    is_flag=True,
    help="print file sizes in human friendly format",
)
@click.option("--long", is_flag=True, help="If set, use a long listing format")
@click.pass_context
def ls(
    ctx: click.Context, path: str, attribs: str, long: bool, raw: bool, human: bool
) -> int:
    # pudb.set_trace()
    ret, resp = asyncio.run(ls_do(options_add(path, attribs, long, ctx.obj["options"])))
    if raw:
        print(resp.message)
        return ret
    lsd: str = resp_process(resp)
    items: str = "\n".join(lsd.split("  "))
    columns_print(filenames_parse(items))
    return ret



import sys
import pudb
from typing import Callable, Optional
from argparse import Namespace
import asyncio
from asyncio import AbstractEventLoop
from pfmongo.models.dataModel import messageType
from pfmongo.config import settings
from pfmisc import Colors as C
from typing import Literal
import os, sys, json
from pathlib import Path
import appdirs
from pfmongo.models import dataModel, responseModel
from typing import Callable, Optional, Any

try:
    from . import __pkg, __version__
except:
    from pfmongo import __pkg, __version__
from tabulate import tabulate
import click
from io import StringIO
from rich.console import Console

NC = C.NO_COLOUR
GR = C.LIGHT_GREEN
CY = C.CYAN
LR = C.LIGHT_RED
LB = C.LIGHT_PURPLE
YL = C.YELLOW


class CustomFormatter(click.HelpFormatter):
    def write(self, string):
        super().write(string)


class CustomGroup(click.Group):
    def format_help(self, ctx, formatter):
        def usage_colorize(usage: str) -> str:
            usage = usage.replace(ctx.command_path, f"{GR}{ctx.command_path}{NC}")
            usage = usage.replace(str(self.name), f"{CY}{self.name}{NC}")
            usage = usage.replace("Usage:", f"{LB}Usage:{NC}")
            return usage

        def options_colorize(options: str) -> str:
            if not "Commands:" in options:
                return options
            l_options: list = options.split("Commands:")[1:][0].split("\n")[1:-1]
            l_cmd: list = [s.split()[0] for s in l_options]
            for cmd in l_cmd:
                options = options.replace(cmd, f"{CY}{cmd}{NC}", 1)
            return options

        # Custom formatting logic for group help
        oformatter: CustomFormatter = CustomFormatter()
        uformatter: CustomFormatter = CustomFormatter()
        super().format_options(ctx, oformatter)
        super().format_usage(ctx, uformatter)
        usage: str = usage_colorize(uformatter.getvalue())
        options: str = options_colorize(oformatter.getvalue())
        click.echo(tabulate_message(str(self.help), usage.strip()))
        click.echo(options)


class CustomCommand(click.Command):
    def format_help(self, ctx, formatter):
        def usage_colorize(usage: str) -> str:
            usage = usage.replace(ctx.command_path, f"{GR}{ctx.command_path}{NC}")
            usage = usage.replace(str(self.name), f"{CY}{self.name}{NC}")
            usage = usage.replace("Usage:", f"{LB}Usage:{NC}")
            return usage

        def options_colorize(options: str) -> str:
            if not "Commands:" in options:
                return options
            l_options: list = options.split("Commands:")[1:][0].split("\n")[1:-1]
            l_cmd: list = [s.split()[0] for s in l_options]
            for cmd in l_cmd:
                options = options.replace(cmd, f"{CY}{cmd}{NC}", 1)
            return options

        # Custom formatting logic for group help
        oformatter: CustomFormatter = CustomFormatter()
        uformatter: CustomFormatter = CustomFormatter()
        super().format_options(ctx, oformatter)
        super().format_usage(ctx, uformatter)
        usage: str = usage_colorize(uformatter.getvalue())
        options: str = options_colorize(oformatter.getvalue())
        click.echo(tabulate_message(str(self.help), usage.strip()))
        click.echo(options)


def tabulate_message(message: str, title: str = "") -> str:
    lines = [[line] for line in message.split("\n")]
    table = tabulate(lines, headers=[f"{title}"], tablefmt="fancy_outline")
    return table


def sprint_multi(string: str, length: int) -> str:
    """
    split a string into several strings of len <length>
    """
    multiline: str = ""
    for i in range(0, len(string), length):
        multiline += string[i : i + length] + "\n"
    return multiline


def complain(message: str, code: int, level: messageType = messageType.INFO) -> int:
    """
    Generate a "complaint" with a message, code, and info level

    :param message: the message
    :param code: the complaint code
    :param level: the type of complaint
    :return: a complaint code
    """
    match level:
        case messageType.ERROR:
            CL = C.RED
        case messageType.INFO:
            CL = C.CYAN
    if settings.appsettings.noComplain:
        return code
    if settings.appsettings.logging == dataModel.loggingType.CONSOLE:
        if message and not settings.appsettings.beQuiet:
            print(tabulate_message(message, f"{CL}{level.name}{NC}"))
    else:
        print(f'{{"level": "{level.name}"}}')
        # pudb.set_trace()
        if message:
            print(f'{{"message": "{to_singleLine(message)}"}}')
    return code


def to_singleLine(message: str) -> str:
    return message.replace("\n", " ")


def URI_sanitize(URI: str) -> str:
    return URI.replace(":", "-").replace("/", "")


def stateDir_get() -> Path:
    return Path(appdirs.user_config_dir(appname=__pkg.name))


def session_config(baseDir: Path) -> Path:
    URI: str = URI_sanitize(settings.mongosettings.MD_URI)
    return baseDir / Path("_MONGO_" + URI)


def sessionUser_notSet() -> int:
    if not settings.mongosettings.MD_sessionUser:
        return complain(
            f"""
                An 'MD_sessionUser' has not been specified in the environment.
                This variable denotes the "name" of a current user of the service
                and is used to store user specific state data.

                Please set with:

                        export {GR}MD_SESSIONUSER{NC}={CY}yourName{NC}
                """,
            5,
            dataModel.messageType.ERROR,
        )
    return 0


def response_exitCode(
    var: (
        responseModel.databaseDesc
        | responseModel.collectionDesc
        | responseModel.showAllDBusage
        | responseModel.showAllcollectionsUsage
        | responseModel.DocumentAddUsage
        | responseModel.DocumentGetUsage
        | responseModel.DocumentDeleteUsage
        | responseModel.DocumentSearchUsage
        | responseModel.DocumentListUsage
        | responseModel.dbDeleteUsage
        | responseModel.CollectionDeleteUsage
    ),
) -> int:
    exitCode: int = 0
    match var:
        case responseModel.databaseDesc() | responseModel.collectionDesc():
            exitCode = 0 if var.info.connected else 100
        case responseModel.showAllDBusage():
            exitCode = 0 if var.status else 101
        case responseModel.showAllcollectionsUsage():
            exitCode = 0 if var.status else 102
        case responseModel.DocumentAddUsage():
            exitCode = 0 if var.status else 103
        case responseModel.DocumentGetUsage():
            exitCode = 0 if var.status else 104
        case responseModel.DocumentDeleteUsage():
            exitCode = 0 if var.status else 105
        case responseModel.DocumentSearchUsage():
            exitCode = 0 if var.status else 106
        case responseModel.DocumentListUsage():
            exitCode = 0 if var.status else 107
        case responseModel.dbDeleteUsage():
            exitCode = 0 if var.status else 108
        case responseModel.CollectionDeleteUsage():
            exitCode = 0 if var.status else 109
    return exitCode


def databaseOrCollectionDesc_message(
    var: responseModel.databaseDesc | responseModel.collectionDesc,
) -> str:
    message: str = ""
    message = (
        f'Successfully connected {var.otype} to "{var.name}"'
        if var.info.connected
        else f'Could not connect to mongo {var.otype}: "{var.name}"'
    )
    return message


def documentAddUsage_message(var: responseModel.DocumentAddUsage) -> str:
    message: str = ""
    name: str = ""
    try:
        name = var.document["_id"]
    except KeyError:
        name = var.documentName
    db: str = var.collection.databaseName
    col: str = var.collection.name
    size: int = 0
    if "_size" in var.document:
        size = var.document["_size"]
    message = (
        f'Successfully added "{name}" (size {size}) to "{db}/{col}"'
        if var.status
        else f'Could not add "{name}" (size {size}) to "{db}/{col}"'
    )
    return message


def documentGetUsage_message(var: responseModel.DocumentGetUsage) -> str:
    # pudb.set_trace()
    message: str = ""
    name: str = var.documentName
    db: str = var.collection.databaseName
    col: str = var.collection.name
    size: int = 0
    if "_size" in var.document:
        size = var.document["_size"]
    message = (
        f"{json.dumps(var.document)}"
        if var.status
        else f'Could not get "{name}" (size {size}) from "{db}/{col}"'
    )
    return message


def documentDeleteUsage_message(var: responseModel.DocumentDeleteUsage) -> str:
    message: str = ""
    name: str = var.documentName
    db: str = var.collection.databaseName
    col: str = var.collection.name
    size: int = 0
    message = (
        f'Successfully deleted "{name}" from "{db}/{col}"'
        if var.status
        else f'Could not delete "{name}" from "{db}/{col}"'
    )
    return message


def documentSearchUsage_message(var: responseModel.DocumentSearchUsage) -> str:
    message: str = ""
    message = f"{var.documentList}" if var.status else ""
    return message


def documentListUsage_message(var: responseModel.DocumentListUsage) -> str:
    message: str = ""
    message = f"{var.documentList}" if var.status else ""
    return message


def showAllDBusage_message(var: responseModel.showAllDBusage) -> str:
    message: str = ""
    message = f"{var.databaseNames}" if var.status else ""
    return message


def showAllcollectionsUsage_message(var: responseModel.showAllcollectionsUsage) -> str:
    message: str = ""
    message = f"{var.collectionNames}" if var.status else ""
    return message


def dbDeleteUsage_message(var: responseModel.dbDeleteUsage) -> str:
    message: str = ""
    message = (
        f'Successfully deleted database "{var.dbName}"'
        if var.status
        else f'Could not delete database "{var.dbName}"'
    )
    return message


def collectionDeleteUsage_message(var: responseModel.CollectionDeleteUsage) -> str:
    message: str = ""
    message = (
        f'Successfully deleted collection "{var.collectionName}"'
        if var.status
        else f'Could not delete collection "{var.collectionName}"'
    )
    return message


def response_messageDesc(
    var: (
        responseModel.databaseDesc
        | responseModel.collectionDesc
        | responseModel.showAllDBusage
        | responseModel.showAllcollectionsUsage
        | responseModel.DocumentAddUsage
        | responseModel.DocumentGetUsage
        | responseModel.DocumentDeleteUsage
        | responseModel.DocumentSearchUsage
        | responseModel.DocumentListUsage
        | responseModel.dbDeleteUsage
        | responseModel.CollectionDeleteUsage
    ),
) -> str:
    message: str = ""
    match var:
        case responseModel.databaseDesc() | responseModel.collectionDesc():
            message = databaseOrCollectionDesc_message(var)
        case responseModel.DocumentAddUsage():
            message = documentAddUsage_message(var)
        case responseModel.DocumentGetUsage():
            message = documentGetUsage_message(var)
        case responseModel.DocumentDeleteUsage():
            message = documentDeleteUsage_message(var)
        case responseModel.DocumentSearchUsage():
            message = documentSearchUsage_message(var)
        case responseModel.showAllDBusage():
            message = showAllDBusage_message(var)
        case responseModel.showAllcollectionsUsage():
            message = showAllcollectionsUsage_message(var)
        case responseModel.DocumentListUsage():
            message = documentListUsage_message(var)
        case responseModel.dbDeleteUsage():
            message = dbDeleteUsage_message(var)
        case responseModel.CollectionDeleteUsage():
            message = collectionDeleteUsage_message(var)
    return message


def connectDB_failureCheck(
    connection: responseModel.databaseDesc,
) -> responseModel.databaseDesc:
    if not connection.info.connected:
        complain(
            f"""
                A database connection error has occured. This typically means
                that the mongo DB service has either not been started or has
                not been specified correctly.

                Please check the service settings. Usually you might just
                need to start the monogo service with:

                        {GR}docker-compose{NC} {CY}up{NC}

                Alternatively, check the mongo service credentials:

                        {GR}export {CY}MD_USERNAME=<user>{NC} && \\
                        {GR}export {CY}MD_PASSWORD=<password>{NC}
                """,
            5,
            dataModel.messageType.ERROR,
        )
    return connection


def connectCollection_failureCheck(
    connection: responseModel.collectionDesc,
) -> responseModel.collectionDesc:
    if not connection.info.connected:
        complain(
            f"""
                A collection connection error has occured. This typically means
                that an connection error was triggered in the housing database.
                Usually this means that the mongo DB service itself has either
                not been started or has not been specified correctly.

                Please check the service settings. Usually you might just
                need to start the monogo service with:

                        {GR}docker-compose{NC} {CY}up{NC}

                Alternatively, check the mongo service credentials:

                        {GR}export {CY}MD_USERNAME=<user>{NC} && \\
                        {GR}export {CY}MD_PASSWORD=<password>{NC}
                """,
            5,
            dataModel.messageType.ERROR,
        )
    return connection


def showAllcollections_failureCheck(
    usage: responseModel.showAllcollectionsUsage,
) -> responseModel.showAllcollectionsUsage:
    if not usage.info.connected:
        complain(
            f"""
                Unable to show all the collections in the database. This typically means
                that the mongo DB service has either not been started or has not been
                specified correctly, or has incorrect credentialling, or other issues.

                Please check the service settings. Usually you might just
                need to start the monogo service with:

                        {GR}docker-compose{NC} {CY}up{NC}
                """,
            5,
            dataModel.messageType.ERROR,
        )
    return usage


def showAllDBUsage_failureCheck(
    usage: responseModel.showAllDBusage,
) -> responseModel.showAllDBusage:
    if not usage.info.connected:
        complain(
            f"""
                Unable to show all databases in the server. This typically means
                that the mongo DB service has either not been started or has not been
                specified correctly, or has incorrect credentialling, or other issues.

                Please check the service settings. Usually you might just
                need to start the monogo service with:

                        {GR}docker-compose{NC} {CY}up{NC}
                """,
            5,
            dataModel.messageType.ERROR,
        )
    return usage


def usage_failureCheck(
    usage: responseModel.databaseNamesUsage | responseModel.collectionNamesUsage,
) -> responseModel.databaseNamesUsage | responseModel.collectionNamesUsage:
    if not usage.info.connected:
        complain(
            f"""
                A usage error has occured. This typically means that the
                mongo DB service has either not been started or has not been
                specified correctly.

                Please check the service settings. Usually you might just
                need to start the monogo service with:

                        {GR}docker-compose{NC} {CY}up{NC}
                """,
            5,
            dataModel.messageType.ERROR,
        )
    return usage


def addDocument_failureCheck(
    usage: responseModel.DocumentAddUsage,
) -> responseModel.DocumentAddUsage:
    # pudb.set_trace()
    if not usage.collection.info.connected:
        complain(
            f"""
                A document add usage error has occured. This typically means that
                the mongo DB service has either not been started or has not been
                specified correctly.

                Please check the service settings. Usually you might just
                need to start the monogo service with:

                        {GR}docker-compose{NC} {CY}up{NC}
                """,
            5,
            dataModel.messageType.ERROR,
        )
    if not usage.status:
        complain(
            f"""
                A document add usage error has occured, reported as:

                {LR}{sprint_multi(usage.resp["error"], 65)}{NC}

                This typically means that a duplicate 'id' has been specified.
                Please check the value of any

                    {CY}--id {GR}<value>{NC}

                 in the {GR}add{NC} subcommand.

                 {CY}{usage.resp["database"]}{NC}/{GR}{usage.resp["collection"]}{NC}/{LR}{usage.documentName}{NC}

                 If this is a "hash" duplication, it means the document content being uploaded  
                 _already_ exists in the database, even if a different 'id' is being specified.
                 If you want to force upload, use a '--noHashing' flag. 

                """,
            6,
            dataModel.messageType.ERROR,
        )

    return usage


def deleteDocument_failureCheck(
    usage: responseModel.DocumentDeleteUsage,
) -> responseModel.DocumentDeleteUsage:
    # pudb.set_trace()
    if not usage.collection.info.connected:
        complain(
            f"""
                A document delete error has occured. This typically means that
                the mongo DB service has either not been started or has not been
                specified correctly.

                Please check the service settings. Usually you might just
                need to start the monogo service with:

                        {GR}docker-compose{NC} {CY}up{NC}
                """,
            5,
            dataModel.messageType.ERROR,
        )
    if not usage.status:
        complain(
            f"""
                A document {YL}delete{NC} usage error has occured. This typically means
                that the passed {CY}id{NC} was not found in the database/collection of
                 {GR}{usage.collection.databaseName}/{usage.collection.name}{NC}

                Please check that

                    {CY}--id {GR}{usage.documentName}{NC}

                in the {YL}delete{NC} subcommand is valid and try again.
                """,
            6,
            dataModel.messageType.ERROR,
        )

    return usage


def deleteDB_failureCheck(
    usage: responseModel.dbDeleteUsage,
) -> responseModel.dbDeleteUsage:
    # pudb.set_trace()
    if not usage.db.info.connected:
        complain(
            f"""
                A db deletion error has occured. This typically means that
                the mongo DB service has either not been started or has not been
                specified correctly.

                Alternatively, the database "{usage.dbName}" does not exist.

                Please check the service settings. Usually you might just
                need to start the monogo service with:

                        {GR}docker-compose{NC} {CY}up{NC}
                """,
            7,
            dataModel.messageType.ERROR,
        )
    elif not usage.status:
        complain(
            f"""
                A database delete usage error has occured. Perhaps the database does not exist?
                """,
            8,
            dataModel.messageType.ERROR,
        )
    return usage


def deleteCollection_failureCheck(
    usage: responseModel.CollectionDeleteUsage,
) -> responseModel.CollectionDeleteUsage:
    # pudb.set_trace()
    if not usage.collection.info.connected:
        complain(
            f"""
                A collection deletion error has occured. This typically means that
                the mongo DB service has either not been started or has not been
                specified correctly.

                Alternatively, the collection "{usage.collectionName}" does not exist.

                Please check the service settings. Usually you might just
                need to start the monogo service with:

                        {GR}docker-compose{NC} {CY}up{NC}
                """,
            7,
            dataModel.messageType.ERROR,
        )
    elif not usage.status:
        complain(
            f"""
                A collection delete usage error has occured in the monogo deletion call.
                """,
            8,
            dataModel.messageType.ERROR,
        )
    return usage


def listDocument_failureCheck(
    usage: responseModel.DocumentListUsage,
) -> responseModel.DocumentListUsage:
    # pudb.set_trace()
    if not usage.collection.info.connected:
        complain(
            f"""
                A document add usage error has occured. This typically means that
                the mongo DB service has either not been started or has not been
                specified correctly.

                Please check the service settings. Usually you might just
                need to start the monogo service with:

                        {GR}docker-compose{NC} {CY}up{NC}
                """,
            7,
            dataModel.messageType.ERROR,
        )
    if not usage.status:
        complain(
            f"""
                A document list usage error has occured. This typically means that
                a non existant search query field has been specified. Please check
                the value of any

                    {CY}--field {GR}<value>{NC}

                 in the {GR}list{NC} subcommand.
                """,
            8,
            dataModel.messageType.ERROR,
        )

    return usage


def getDocument_failureCheck(
    usage: responseModel.DocumentGetUsage,
) -> responseModel.DocumentGetUsage:
    # pudb.set_trace()
    if not usage.collection.info.connected:
        complain(
            f"""
                A document add usage error has occured. This typically means that
                the mongo DB service has either not been started or has not been
                specified correctly.

                Please check the service settings. Usually you might just
                need to start the monogo service with:

                        {GR}docker-compose{NC} {CY}up{NC}
                """,
            7,
            dataModel.messageType.ERROR,
        )
    if not usage.status:
        complain(
            f"""
                A document list get error has occured. This typically means that
                no document in the collection had a matching id term. Please check
                the value of any

                    {CY}--id {GR}<value>{NC}

                 in the {GR}get{NC} subcommand.
                """,
            8,
            dataModel.messageType.ERROR,
        )

    return usage


def searchDocument_failureCheck(
    usage: responseModel.DocumentSearchUsage,
) -> responseModel.DocumentSearchUsage:
    # pudb.set_trace()
    if not usage.collection.info.connected:
        complain(
            f"""
                A document add usage error has occured. This typically means that
                the mongo DB service has either not been started or has not been
                specified correctly.

                Please check the service settings. Usually you might just
                need to start the monogo service with:

                        {GR}docker-compose{NC} {CY}up{NC}
                """,
            7,
            dataModel.messageType.ERROR,
        )
    if not usage.status:
        complain(
            f"""
                A document search usage error has occured. This typically means that
                a non existant search query field has been specified. Please check
                the value of any

                    {CY}--field {GR}<value>{NC}

                 in the {GR}search{NC} subcommand.
                """,
            8,
            dataModel.messageType.ERROR,
        )

    return usage


def env_statePathSet(options: Namespace) -> bool:
    """
    Check/set a path to contain state/persistent data, typically the
    database name and collection within that database.

    The value is set in the options.statePath, and the return value
    indicates failure/success.

    :param options: the set of options
    :return: 0 -- success, non-zero -- failure
    """
    options.thisSession = session_config(stateDir_get())
    if sessionUser_notSet():
        return False
    options.sessionUser = settings.mongosettings.MD_sessionUser
    options.statePath = options.thisSession / options.sessionUser
    if not options.statePath.exists():
        options.statePath.mkdir(parents=True, exist_ok=True)
    return True


def DB_stateFileResolve(options: Namespace) -> Path:
    return options.statePath / Path(settings.mongosettings.stateDBname)


def collection_stateFileResolve(options: Namespace) -> Path:
    return options.statePath / Path(settings.mongosettings.stateColname)


def DBname_stateFileRead(options: Namespace) -> str:
    contents: str = ""
    statefile: Path = DB_stateFileResolve(options)
    if statefile.exists():
        contents = statefile.read_text()
    return contents


def DBname_stateFileSave(options: Namespace, contents: str) -> str:
    statefile: Path = DB_stateFileResolve(options)
    statefile.write_text(contents)
    return contents


def collectionName_stateFileRead(options: Namespace) -> str:
    contents: str = ""
    statefile: Path = collection_stateFileResolve(options)
    if statefile.exists():
        contents = statefile.read_text()
    return contents


def collectionName_stateFileSave(options: Namespace, contents: str) -> str:
    statefile: Path = collection_stateFileResolve(options)
    statefile.write_text(contents)
    return contents


def stateFileSave(
    options: Namespace, contents: str, f_stateFileResolve: Callable[[Namespace], Path]
) -> str:
    statefile: Path = f_stateFileResolve(options)
    statefile.write_text(contents)
    return contents


def stateFileRead(
    options: Namespace, f_stateFileResolve: Callable[[Namespace], Path]
) -> str:
    contents: str = ""
    statefile: Path = f_stateFileResolve(options)
    if statefile.exists():
        contents = statefile.read_text()
    return contents


def collectionName_get(options: Namespace) -> str:
    """
    Determine the name of the collection within the mongo server to use.

    Order of precedence:
        * if '--useCollection' (collectionName) in args, use this as
          the collection and set the same value in the settings object;
        * if no '--useCollection', check for a state file in the options.statePath
          and if this exists, read that file and set the collectionName and
          settings object;
        * if neither, then check the settings object and set the collectionName
          to that;
        * otherwise, failing everything, return an empty string.

    :param options: the set of CLI (and more) options
    :return: a string database name
    """
    if options.collectionName:
        settings.mongosettings.MD_COLLECTION = options.collectionName
        collectionName_stateFileSave(options, options.collectionName)
    if not options.collectionName:
        collectionName: str = collectionName_stateFileRead(options)
        if collectionName:
            options.collectionName = collectionName
            settings.mongosettings.MD_COLLECTION = options.DBname
    # if not options.collectionName:
    #     options.collectionName  = settings.mongosettings.MD_COLLECTION
    return options.collectionName


def DBname_get(options: Namespace) -> str:
    """
    Determine the name of the database within the mongo server to use.

    Order of precedence:
        * if '--useDB' (i.e. DBname) in args, use this as the DBname
          and set the same value in the settings object;
        * if no '--DBname', check for a state file in the options.statePath
          and if this exists, read that file and set the DBname and settings
          object;
        * if neither, then check the settings object and set the DBname to
          that;
        * otherwise, failing everything, return an empty string.

    :param options: the set of CLI (and more) options
    :return: a string database name
    """
    if options.DBname:
        settings.mongosettings.MD_DB = options.DBname
        DBname_stateFileSave(options, options.DBname)
    if not options.DBname:
        DBname: str = DBname_stateFileRead(options)
        if DBname:
            options.DBname = DBname
            settings.mongosettings.MD_DB = options.DBname
    # if not options.DBname:
    #     options.DBname                  = settings.mongosettings.MD_DB
    return options.DBname


def env_failCheck(options: Namespace) -> int:
    if "--help" in sys.argv:
        return 0
    if not DBname_get(options):
        return complain(
            f"""
            Unable to determine which database to use.

            No resolution method was successful. Resolutions include:
                * a {GR}--useDB {YL}<database>{NC} CLI key/value pair
                * an appropriate MD_DB environment variable
                * a previous CLI {YL}database{NC} subcommand
                * an appropriate CLI {CY}cd {YL}/<database>{NC}
            """,
            1,
            messageType.ERROR,
        )
    if not collectionName_get(options):
        return complain(
            f"""
            Unable to determine the collection within
            the database {C.YELLOW}{DBname_get(options)}{NC} to use.

            A `--useCollection` flag with the collection
            as argument must be specified or alternatively
            be set in the environment as MD_COLLECTION or
            exist as a previous configuration state. """,
            2,
            messageType.ERROR,
        )
    return 0



import  click
from    pathlib import Path
from    pfmisc  import Colors as C

NC  = C.NO_COLOUR
GR  = C.GREEN
CY  = C.CYAN
YL  = C.YELLOW

@click.command(help=f"""
export a {YL}document{NC} to a {CY}file{NC}

The {GR}exp{NC} command "exports" a document {CY}pathIN{NC} to the real filesystem
at {GR}pathEX{NC}.


""")
@click.argument('pathIN',
                required = True)
@click.argument('pathEX',
                required = True)
def exp(path:str) -> None:
    # pudb.set_trace()
    target:Path     = Path('')
    if path:
        target = Path(path)





from pfmongo import pfmongo
from pfmongo.__main__ import main
import os, pudb
from pfmongo.commands.dbop import connect as database, deleteDB
from pfmongo.commands.clop import connect as collection
from pfmongo.commands.docop import add, get
from pfmongo.models import responseModel
from pfmongo.config import settings
import json
import pytest
import asyncio
import uuid

os.environ["XDG_CONFIG_HOME"] = "/tmp"


@pytest.fixture(autouse=True)
async def cleanup():
    """
    Cleanup databases after each test function in this file.
    """
    await DB_delete()
    yield
    await DB_delete()  # Clean up the database after the test runs


@pytest.mark.asyncio
async def DB_connect(DB: str = "testDB") -> int:
    """
    Connect to the specified MongoDB database.
    """
    return await database.connectTo_asInt(
        database.options_add(DB, pfmongo.options_initialize())
    )


@pytest.mark.asyncio
async def collection_connect(col: str = "testCollection") -> int:
    """
    Connect to the specified MongoDB collection.
    """
    return await collection.connectTo_asInt(
        collection.options_add(col, pfmongo.options_initialize())
    )


@pytest.mark.asyncio
async def DB_delete(DB: str = "testDB") -> int:
    """
    Delete the specified MongoDB database.
    """
    return await deleteDB.DBdel_asInt(
        deleteDB.options_add(DB, pfmongo.options_initialize())
    )


@pytest.mark.asyncio
async def test_duplicateHash_add_asModel() -> None:
    """
    Test handling duplicate document hashes.
    """
    await DB_delete()
    await DB_connect()
    await collection_connect()
    # pudb.set_trace()
    retlld: responseModel.mongodbResponse = await add.documentAdd_asModel(
        add.options_add("examples/lld.json", "lld2.json", pfmongo.options_initialize())
    )
    retlld = await add.documentAdd_asModel(
        add.options_add("examples/lld.json", "lld3.json", pfmongo.options_initialize())
    )
    print(retlld.response)
    print(retlld.response["connect"].resp)
    print(settings.appsettings)
    assert not retlld.response["status"]
    assert "Duplicate document hash found." in retlld.response["connect"].resp["error"]
    await DB_delete()


if __name__ == "__main__":
    print("Test document duplicate hash detection")
    pudb.set_trace()
    asyncio.run(test_duplicateHash_add_asModel())



str_description = """

    Some data models/schemas.

"""

from    pydantic            import BaseModel, Field
from    typing              import Optional, List, Dict
from    datetime            import datetime
from    enum                import Enum
from    pathlib             import Path

class cdResponse(BaseModel):
    path:Path       = Path("")
    message:str     = ""
    status:bool     = True
    error:str       = ""
    code:int        = 0
    state:dict      = {
                        "database":     "",
                        "collection":   ""
                       }





# Turn off all logging for modules in this libary.
import logging

logging.disable(logging.CRITICAL)

# System imports
import sys
import logging
import datetime

from argparse import Namespace, ArgumentParser
from argparse import RawTextHelpFormatter
from loguru import logger

from pfmisc import Colors as C
import pudb
from typing import Any, Callable, Optional

from pfmongo.config import settings
from pfmongo.db import pfdb
from pfmongo.models import responseModel
from pfmongo import env
import copy

NC = C.NO_COLOUR
YL = C.YELLOW
GR = C.GREEN
CY = C.CYAN
PR = C.PURPLE
LOG = logger.debug
logger_format = (
    "<green>{time:YYYY-MM-DD HH:mm:ss}</green> │ "
    "<level>{level: <5}</level> │ "
    "<yellow>{name: >28}</yellow>::"
    "<cyan>{function: <30}</cyan> @"
    "<cyan>{line: <4}</cyan> ║ "
    "<level>{message}</level>"
)
logger.remove()
logger.add(sys.stderr, format=logger_format)

package_description: str = f"""
                        {YL}pfmongo{NC} -- mongo for rest of us.

    This system is a somewhat ideosyncratic interface to a {PR}mongodb{NC} server.
    It is both a CLI client and supporting python module, following the "pf"
    (see elsewhere) design pattern.

    The basic idea is to provide a subcommand based set of operations to
    simplify working with data (documents) in a {PR}mongodb{NC}.

    For added happiness, a "file system-y" metaphor complete with supporting
    commands that present the database as a file system is also provided.

    In addtion, see the {PR}pfmdb{NC} FastAPI-based system that provides a full
    REST interface to {YL}pfmongo{NC}.

    From the CLI, subcommands are organized the following groupings:

        {GR}fs{NC}          for 'filesystem' type commands,
        {GR}database{NC}    for 'database' type commands,
        {GR}collection{NC}  for 'collection' type commands,
        {GR}document{NC}    for 'document' type commands
        {GR}smash{NC}       the pfmongo REPL 'shell'

    Use {YL}pfmongo {GR}<grouping> {CY}--help{NC} for <grouping> specific help. For example

        {YL}pfmongo {GR}document {CY}--help{NC}

    for help on the {GR}document{NC} contextual commands.

"""

package_coreDescription: str = f"""
    {YL}pfmongo{NC} supports some core CLI arguments that are used to specify
    the general operation of the program, specifically the mongo database to
    use and the collection within that database to access.

    These can also be set with CLI and are specified _before_ the sub command
    to use.

    Additionally, a simple --version can also be specified here.
"""

package_CLIself = """
        [--useDB <DBname>]                                                      \\
        [--useCollection <collectionName>]                                      \\
        [--noHashing]                                                           \\
        [--noDuplicates]                                                        \\
        [--donotFlatten]                                                        \\
        [--noResponseTruncSize]                                                 \\
        [--detailedOutput]                                                      \\
        [--noComplain]                                                          \\
        [--beQuiet]                                                             \\
        [--eventLoopDebug]                                                      \\
        [--responseTruncDepth <depth>]                                          \\
        [--responseTruncSize <size>]                                            \\
        [--man]                                                                 \\
        [--version]"""

package_argSynopsisSelf = f"""
        {YL}[--useDB <DBname>]{NC}
        Use the data base called <DBname>.

        {YL}[--useCollection <collectionName>]{NC}
        Use the collection called <collectionName>.

        {YL}[--noHashing>]{NC}
        If set, do not add payload hashes to documents. By default a hash of
        the document is calculated and added as a field when saved to a
        collection. This hash serves as a "fingerprint" of the document
        assuring _content_ uniqueness.

        {YL}[--noDuplicates]{NC}
        If set, and in combination with --noHashing, will not allow a new
        document with the same hash as an existing one to be saved to a
        collection.

        {YL}[--donotFlatten]{NC}
        If set, do not add a flattened document to a shadow collection. Since
        the interal structure of a document can be complex with nested data
        structures, a "flattened" version of the document is added to a
        "shadow" collection. This flattened version is consulted for quick
        value searching. In other words, flattened documents are internal
        artifacts and a side-effect of adding data to the collection. By
        passing this flag, a flattened version is not created or saved. Note
        that not flattening will severely limit searches!

        {YL}[--noResponseSizeTrunc]{NC}
        If set, do not truncate response returns from pfmongo calls. Note that
        some responses can be "wordy", so only use if strictly needed.

        {YL}[--noComplain]{NC}
        If set, suppress the "complain" messages that typically provide more
        information on error or warning events.

        {YL}[--beQuiet]{NC}
        If set, suppress all output to console. Note that "models" still contain
        data and the contents of the model '.message' is the string that would
        typically have been printed. This option is mostly useful if some other
        process is using pfmongo and wants to prevent pfmongo itself from polluting
        the console.

        {YL}[--detailedOutput]{NC}
        If set, more detailed output result payloads.

        {YL}[--eventLoopDebug]{NC}
        If set, activate a hidden "pudb.set_trace()" in the main event loop.

        {YL}[--responseTruncSize <size>]{NC}
        Set the size limit on a given (nested) set of data in the response from
        the app. You probably won'y need this.

        {YL}[--responseTruncDepth <depth>]{NC}
        Set the depth within a response below which to possibly truncatate on
        response size. You probably won't need this.

        {YL}[--man]{NC}
        If specified, print this help/man page.

        {YL}[--version]{NC}
        If specified, print app name/version.
"""

package_CLIfull = package_CLIself
package_CLIDS = package_CLIself
package_argsSynopsisFull = package_argSynopsisSelf
package_argsSynopsisDS = package_argSynopsisSelf


def parser_setup(desc: str, add_help: bool = True) -> ArgumentParser:
    parserSelf = ArgumentParser(
        description=desc, formatter_class=RawTextHelpFormatter, add_help=add_help
    )

    parserSelf.add_argument(
        "--monogdbinit",
        help="JSON formatted file containing mongodb initialization",
        dest="mongodbinit",
        default="",
    )

    parserSelf.add_argument(
        "--noHashing",
        help="do not add hashes to inserted documents",
        dest="noHashing",
        default=settings.appsettings.noHashing,
        action="store_true",
    )

    parserSelf.add_argument(
        "--allowDuplicates",
        help="allow duplicate (hashed) documents",
        dest="allowDuplicates",
        default=settings.appsettings.allowDuplicates,
        action="store_true",
    )

    parserSelf.add_argument(
        "--donotFlatten",
        help="do not add shadow 'flattened' collection",
        dest="donotFlatten",
        default=settings.appsettings.donotFlatten,
        action="store_true",
    )

    parserSelf.add_argument(
        "--noResponseTruncSize",
        help="do not truncate responses, even if nested and high data",
        dest="noResponseTruncSize",
        default=settings.appsettings.noResponseTruncSize,
        action="store_true",
    )

    parserSelf.add_argument(
        "--noComplain",
        help="suppress complaint details",
        dest="noComplain",
        default=settings.appsettings.noComplain,
        action="store_true",
    )

    parserSelf.add_argument(
        "--beQuiet",
        help="suppress all console output",
        dest="beQuiet",
        default=settings.appsettings.beQuiet,
        action="store_true",
    )

    parserSelf.add_argument(
        "--detailedOutput",
        help="provide more detailed result payloads",
        dest="detailedOutput",
        default=settings.appsettings.detailedOutput,
        action="store_true",
    )

    parserSelf.add_argument(
        "--eventLoopDebug",
        help="activate a hidden event loop breakpoint",
        dest="eventLoopDebug",
        default=settings.appsettings.eventLoopDebug,
        action="store_true",
    )

    parserSelf.add_argument(
        "--modelSizesPrint",
        help="on final response, also print internal model size information",
        dest="modelSizesPrint",
        default=settings.appsettings.modelSizesPrint,
        action="store_true",
    )

    parserSelf.add_argument(
        "--responseTruncSize",
        help="set the cumulative truncation size limit",
        dest="responseTruncSize",
        default="",
    )

    parserSelf.add_argument(
        "--responseTruncDepth",
        help="set the nested depth at which to consider truncation",
        dest="responseTruncDepth",
        default="",
    )

    parserSelf.add_argument(
        "--version",
        help="print name and version",
        dest="version",
        default=False,
        action="store_true",
    )

    parserSelf.add_argument(
        "--man", help="print man page", dest="man", default=False, action="store_true"
    )

    parserSelf.add_argument(
        "--useDB", help="use the named data base", dest="DBname", default=""
    )

    parserSelf.add_argument(
        "--useCollection",
        help="use the named collection",
        dest="collectionName",
        default="",
    )

    return parserSelf


def parser_interpret(parser, args) -> tuple:
    """
    Interpret the list space of *args, or sys.argv[1:] if
    *args is empty
    """
    if args:
        args, unknown = parser.parse_known_args(args)
    else:
        args, unknown = parser.parse_known_args(sys.argv[1:])
    return args, unknown


def parser_JSONinterpret(parser, d_JSONargs) -> tuple:
    """
    Interpret a JSON dictionary in lieu of CLI.

    For each <key>:<value> in the d_JSONargs, append to
    list two strings ["--<key>", "<value>"] and then
    argparse.
    """
    l_args = []
    for k, v in d_JSONargs.items():
        if v.isinstance(bool):
            if v:
                l_args.append("--%s" % k)
            continue
        l_args.append("--%s" % k)
        l_args.append("%s" % v)
    return parser_interpret(parser, l_args)


def options_initialize(override: list[dict] = []) -> Namespace:
    """initialize a set of "options" with defaults"""
    options: Namespace = Namespace()
    parser: ArgumentParser = parser_setup("A client parser for pfmongo", add_help=False)
    options, extra = parser_interpret(parser, [])
    if len(override):
        options = options_set(options, override)
    env.env_statePathSet(options)
    return options


def timenow(fmt: str = "%Y-%m-%d %H:%M:%S") -> str:
    return datetime.datetime.now().strftime(fmt)


def date_toUNIX(str_date: str) -> int:
    ret: int = 0
    try:
        date_obj: datetime.datetime = datetime.datetime.strptime(str_date, "%Y-%m-%d")
        ret = int(date_obj.timestamp())
    except Exception as e:
        print(f"{e}")
    return ret


def responseData_build(
    d_mongoresp: dict, message: str = ""
) -> responseModel.mongodbResponse:
    d_resp: responseModel.mongodbResponse = responseModel.mongodbResponse()
    d_resp.status = d_mongoresp["status"]
    d_resp.response = d_mongoresp
    d_resp.message = message
    return d_resp


def options_set(options: Namespace, newkeyvaluepair: list) -> Namespace:
    """
    set some values (newkeyvaluepair) in a copy of
    <options> and return this copy. Note the original
    <options> is not affected.
    """
    localoptions = copy.deepcopy(options)
    for pair in newkeyvaluepair:
        for k, v in pair.items():
            setattr(localoptions, k, v)
    return localoptions


class Pfmongo:
    """

    A class that provides a python API for interacting with a mongodb.

    """

    def setup_fromCLI(self) -> None:
        settings.appsettings.allowDuplicates = self.args.allowDuplicates
        settings.appsettings.noHashing = self.args.noHashing
        settings.appsettings.donotFlatten = self.args.donotFlatten
        settings.appsettings.detailedOutput = self.args.detailedOutput
        settings.appsettings.noComplain = self.args.noComplain
        settings.appsettings.eventLoopDebug = self.args.eventLoopDebug
        settings.appsettings.noResponseTruncSize = self.args.noResponseTruncSize
        settings.appsettings.modelSizesPrint = self.args.modelSizesPrint
        settings.appsettings.beQuiet = self.args.beQuiet
        if self.args.responseTruncSize:
            settings.mongosettings.responseTruncSize = self.args.responseTruncSize
        if self.args.responseTruncDepth:
            settings.mongosettings.responseTruncDepth = self.args.responseTruncDepth

    def __init__(self, args, **kwargs) -> None:
        self.args: Namespace = Namespace()
        if isinstance(args, dict):
            parser: ArgumentParser = parser_setup("Setup client using dict")
            self.args, extra = parser_JSONinterpret(parser, args)
        if type(args) is Namespace:
            self.args = args

        # attach a comms API to the mongo db
        self.dbAPI: pfdb.mongoDB = pfdb.mongoDB(
            settings=settings.mongosettings, args=self.args
        )

        self.responseData: responseModel.mongodbResponse = (
            responseModel.mongodbResponse()
        )
        self.exitCode: int = 1
        self.setup_fromCLI()

    def responseData_log(self, data: Any) -> responseModel.mongodbResponse:
        self.exitCode = env.response_exitCode(data)
        self.responseData = responseData_build(
            {"status": data.status, "connect": data}, env.response_messageDesc(data)
        )
        return self.responseData

    async def showAllDB(self) -> responseModel.showAllDBusage:
        self.responseData = self.responseData_log(
            (allDB := await self.dbAPI.database_names_get())
        )
        return allDB

    async def showAllCollections(self) -> responseModel.showAllcollectionsUsage:
        allCollections: responseModel.showAllcollectionsUsage = (
            responseModel.showAllcollectionsUsage()
        )
        if not (
            connectDB := await self.connectDB(env.DBname_get(self.args))
        ).info.connected:
            allCollections.info = connectDB.info
            return allCollections
        allCollections = await self.dbAPI.collection_names_get()
        allCollections.info = connectDB.info
        self.responseData = self.responseData_log(allCollections)
        return allCollections

    async def deleteCollection(
        self, collection: str
    ) -> responseModel.CollectionDeleteUsage:
        collectionDelete: responseModel.CollectionDeleteUsage = (
            responseModel.CollectionDeleteUsage()
        )
        if not (
            connectCol := await self.connectCollection_do(
                env.collectionName_get(self.args)
            )
        ).info.connected:
            collectionDelete.collection = connectCol
            return collectionDelete
        self.responseData_log(
            collectionDelete := await self.dbAPI.collection_delete(collection)
        )
        return collectionDelete

    async def documentAdd(self, d_json: dict) -> responseModel.DocumentAddUsage:
        documentAdd: responseModel.DocumentAddUsage = responseModel.DocumentAddUsage()
        if not (
            connectCol := await self.connectCollection_do(
                env.collectionName_get(self.args)
            )
        ).info.connected:
            documentAdd.collection = connectCol
            return documentAdd
        self.responseData_log(
            documentAdd := await self.dbAPI.insert_one(connectCol, document=d_json)
        )
        return documentAdd

    async def documentDelete(self, id: str) -> responseModel.DocumentDeleteUsage:
        documentDelete: responseModel.DocumentDeleteUsage = (
            responseModel.DocumentDeleteUsage()
        )
        if not (
            connectCol := await self.connectCollection_do(
                env.collectionName_get(self.args)
            )
        ).info.connected:
            documentDelete.collection = connectCol
            return documentDelete
        self.responseData_log(
            documentDelete := await self.dbAPI.delete_one(connectCol, document=id)
        )
        return documentDelete

    async def documentList(self, field: str) -> responseModel.DocumentListUsage:
        documentList: responseModel.DocumentListUsage = (
            responseModel.DocumentListUsage()
        )
        if not (
            connectCol := await self.connectCollection_do(
                env.collectionName_get(self.args)
            )
        ).info.connected:
            documentList.collection = connectCol
            return documentList
        self.responseData_log(
            documentList := await self.dbAPI.listDocs(connectCol, field=field)
        )
        return documentList

    async def documentSearch(
        self, searchFor: list, field: str
    ) -> responseModel.DocumentSearchUsage:
        """search whatever is the contextual collection!"""
        documentList: responseModel.DocumentSearchUsage = (
            responseModel.DocumentSearchUsage()
        )
        collection: str = env.collectionName_get(self.args)
        # if not settings.appsettings.donotFlatten:
        #     collection += settings.mongosettings.flattenSuffix
        if not (
            connectCol := await self.connectCollection_do(collection)
        ).info.connected:
            documentList.collection = connectCol
            return documentList
        self.responseData_log(
            documentList := await self.dbAPI.searchDocs(
                connectCol, searchFor=searchFor, field=field
            )
        )
        # reconnect to non-flat in a thread safe way?
        return documentList

    async def documentGet(self, id: str) -> responseModel.DocumentGetUsage:
        documentGet: responseModel.DocumentGetUsage = responseModel.DocumentGetUsage()
        if not (
            connectCol := await self.connectCollection_do(
                env.collectionName_get(self.args)
            )
        ).info.connected:
            documentGet.collection = connectCol
            return documentGet
        self.responseData_log(
            documentGet := await self.dbAPI.get_one(connectCol, document=id)
        )
        return documentGet

    async def connectDB(self, DBname: str) -> responseModel.databaseDesc:
        connect: responseModel.databaseDesc = responseModel.databaseDesc()
        self.responseData_log((connect := await self.dbAPI.connect(DBname)))
        if self.responseData.response["status"]:
            env.stateFileSave(self.args, DBname, env.DB_stateFileResolve)
        return connect

    async def deleteDB(self, DB: str) -> responseModel.dbDeleteUsage:
        dbDelete: responseModel.dbDeleteUsage = responseModel.dbDeleteUsage()
        if not (
            connectDB := await self.connectDB_do(env.DBname_get(self.args))
        ).info.connected:
            dbDelete.db = connectDB
            return dbDelete
        dbDelete.dbName = DB
        dbDelete.db = connectDB
        self.responseData_log(dbDelete := await self.dbAPI.db_delete(dbDelete))
        return dbDelete

    async def connectCollection(
        self, collectionName: str
    ) -> responseModel.collectionDesc:
        connectCol: responseModel.collectionDesc = responseModel.collectionDesc()
        connectDB: responseModel.databaseDesc = await self.connectDB(
            env.DBname_get(self.args)
        )
        connectCol.database = connectDB
        if not connectDB.info.connected:
            return connectCol
        connectCol = await self.dbAPI.collection_connect(collectionName)
        connectCol.database = connectDB
        self.responseData = self.responseData_log(connectCol)
        # pudb.set_trace()
        if self.responseData.response["status"]:
            env.stateFileSave(
                self.args, collectionName, env.collection_stateFileResolve
            )
            self.args.collectionName = collectionName
        return connectCol

    async def collectionDesc_check(self) -> responseModel.collectionDesc:
        connectCol: responseModel.collectionDesc = responseModel.collectionDesc()
        DBdesc: responseModel.databaseDesc = await self.connectDB(
            env.DBname_get(self.args)
        )

        if not DBdesc.info.connected:
            connectCol.info.error = DBdesc.info.error
            return connectCol
        connectCol = await self.dbAPI.collection_connect(
            env.collectionName_get(self.args)
        )
        self.responseData = self.responseData_log(connectCol)
        if self.responseData.response["status"]:
            env.stateFileSave(
                self.args,
                env.collectionName_get(self.args),
                env.collection_stateFileResolve,
            )
        return connectCol

    async def showAllDB_do(self) -> responseModel.showAllDBusage:
        return env.showAllDBUsage_failureCheck(await self.showAllDB())

    async def showAllCollections_do(self) -> responseModel.showAllcollectionsUsage:
        return env.showAllcollections_failureCheck(await self.showAllCollections())

    async def connectDB_do(self, DBname: str) -> responseModel.databaseDesc:
        return env.connectDB_failureCheck(await self.connectDB(DBname))

    async def connectCollection_do(
        self, collection: str
    ) -> responseModel.collectionDesc:
        return env.connectCollection_failureCheck(
            await self.connectCollection(collection)
        )

    async def addDocument_do(self, document: dict) -> responseModel.DocumentAddUsage:
        return env.addDocument_failureCheck(await self.documentAdd(document))

    async def deleteDocument_do(self, id: str) -> responseModel.DocumentDeleteUsage:
        return env.deleteDocument_failureCheck(await self.documentDelete(id))

    async def deleteCollection_do(
        self, collection: str
    ) -> responseModel.CollectionDeleteUsage:
        return env.deleteCollection_failureCheck(
            await self.deleteCollection(collection)
        )

    async def deleteDB_do(self, DB: str) -> responseModel.dbDeleteUsage:
        return env.deleteDB_failureCheck(await self.deleteDB(DB))

    async def listDocument_do(self, field: str) -> responseModel.DocumentListUsage:
        return env.listDocument_failureCheck(await self.documentList(field))

    async def getDocument_do(self, id: str) -> responseModel.DocumentGetUsage:
        return env.getDocument_failureCheck(await self.documentGet(id))

    async def searchDocument_do(
        self, search: dict
    ) -> responseModel.DocumentSearchUsage:
        return env.searchDocument_failureCheck(
            await self.documentSearch(search["searchFor"], search["field"])
        )

    async def service(self) -> None:
        if settings.appsettings.eventLoopDebug:
            pudb.set_trace()

        if not hasattr(self.args, "do"):
            return

        match self.args.do:
            case "showAllDB":
                await self.showAllDB_do()
            case "showAllCollections":
                await self.showAllCollections_do()
            case "connectDB":
                await self.connectDB_do(self.args.argument)
            case "deleteDB":
                await self.deleteDB_do(self.args.argument)
            case "connectCollection":
                await self.connectCollection_do(self.args.argument)
            case "addDocument":
                await self.addDocument_do(self.args.argument)
            case "deleteDocument":
                await self.deleteDocument_do(self.args.argument)
            case "deleteCollection":
                await self.deleteCollection_do(self.args.argument)
            case "listDocument":
                await self.listDocument_do(self.args.argument)
            case "getDocument":
                await self.getDocument_do(self.args.argument)
            case "searchDocument":
                await self.searchDocument_do(self.args.argument)
